/*
 *
 * Copyright 2017-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import static de.bsvrz.ars.ars.persistence.ContainerFile.BUFFER_SIZE;
import static de.bsvrz.ars.ars.persistence.ContainerFile.DATALEN_LEN;
import static de.bsvrz.ars.ars.persistence.ContainerFile.MAX_UNCOMPRESSED;
import static de.bsvrz.ars.ars.persistence.ContainerFile.NOT_COMPRESSED;
import static de.bsvrz.ars.ars.persistence.ContainerFile.NO_DATA;
import static de.bsvrz.ars.ars.persistence.ContainerFile.NO_RIGHTS;
import static de.bsvrz.ars.ars.persistence.ContainerFile.NO_SOURCE;
import static de.bsvrz.ars.ars.persistence.ContainerFile.POT_GAP;
import static de.bsvrz.ars.ars.persistence.ContainerFile.getTotalDataOverhead;


import de.bsvrz.ars.ars.persistence.ContainerFile;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.archive.ArchiveData;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.sys.funclib.dataSerializer.Serializer;
import de.bsvrz.sys.funclib.dataSerializer.SerializingFactory;
import de.bsvrz.sys.funclib.losb.util.ByteIO;
import de.bsvrz.sys.funclib.losb.util.Util;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.zip.Deflater;

/**
 * Klasse, die Archivdaten serialisiert um diese in Containerdateien zu schreiben.
 *
 * @author Kappich Systemberatung
 * @see de.bsvrz.ars.ars.persistence.ContainerFileHandle#appendSerializedData(byte[], int)
 */
public class ArchiveDataSerializer {

	/**
	 * Referenz auf den tatsaechlichen, von {@link #serializeData(ResultData, long, Deflater)} verwendeten Puffer, der von {@link #writeBuf}
	 * abweicht, falls dieser zu klein ist). Notwendig, da Daten vorserialisiert werden müssen, um deren Länge zu bestimmen.
	 */
	private byte[] actualWriteBuf;

	/** Länge des von {@link #serializeData(ResultData, long, Deflater)} serialisierten Datensatzes. */
	private int totalWriteDataSize;

	/** Puffer zum Kopieren von Dateien oder Einlesen von Strings und Zahlen. Wird nur von ArchivTasks benutzt (und damit automatisch angelegt) */
	private byte[] writeBuf;

	/** Puffer für den Serialisierer und zum Lesen eines Datensatzes. */
	private ByteArrayOutputStream bos;

	/** Datensatz-Serialisierer. */
	private Serializer serializer;

	/** Legt alle zum Schreiben benoetigte Puffer an. Wird nur von Schreib-Methoden aufgerufen. Lesende Tasks beanspruchen so erheblich weniger Speicher. */
	private void createWriteBuffers() {
		// Nur anlegen, falls schreibender Task (ArchiveTask) zugreift
		if(bos == null) bos = new ByteArrayOutputStream(BUFFER_SIZE);
		if(writeBuf == null) writeBuf = new byte[BUFFER_SIZE];
		if(serializer == null) serializer = SerializingFactory.createSerializer(bos);
	}

	/**
	 * Serialisiert den übergebenen Datensatz und speichert ihn in {@link #actualWriteBuf} und {@link #totalWriteDataSize}. Da der
	 * ArchivTask vor dem Archivieren wissen muss, ob der Container zu lang wird und abgeschlossen werden muss, ist das Anhängen eines Datensatzes auf diese
	 * Methode und {@link ContainerFile#appendSerializedData(ArchiveDataSerializer)} aufgeteilt. Wenn der Datensatz ein
	 * Spezialfall ist (keine Quelle, keine Daten, Datenlücke), wird ein speziell codierter Datensatz verwendet.
	 *
	 * @param rd         Zu archivierender Datensatz
	 * @param arcTime    Archivzeitstempel
	 * @param compressor Komprimierungsklasse
	 * @return Länge des serialisierten Datensatzes
	 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
	 */
	public int serializeData(ResultData rd, long arcTime, Deflater compressor) throws PersistenceException {
		createWriteBuffers();    // Puffer (nur für schreibende Tasks) lazy anlegen

		byte[] data;
		if(rd.hasData()) {
			// normaler Datensatz
			bos.reset();
			try {
				serializer.writeData(rd.getData());
			}
			catch(IOException e) {
				throw new PersistenceException("Datensatz konnte nicht serialisiert werden: " + e.getMessage() + "\n" + rd, e);
			}
			data = bos.toByteArray();
		}
		else {
			// kuenstlicher Datensatz (keine Quelle, keine Daten)
			data = writeDataModifier(rd);
		}
		actualWriteBuf = getBuf(writeBuf, getTotalDataOverhead() + data.length);    // Nach Möglichkeit bereits angelegten Puffer verwenden
		return embedSerialializedData(arcTime, rd.getDataTime(), rd.getDataIndex(), data, actualWriteBuf, compressor);
	}

	/**
	 * Liefert den übergebenen Puffer falls dessen Größe ausreichend ist, andernfalls einen neu angelegten. Zweck: sowenig Arbeit für die Speicherverwaltung
	 * wie möglich.
	 *
	 * @param defaultBuffer Vorhandener Puffer
	 * @param desiredSize   Gewuenschte PufferGröße
	 * @return Byte-Array-Puffer
	 */
	private static byte[] getBuf(byte[] defaultBuffer, int desiredSize) {
		return (desiredSize <= defaultBuffer.length) ? defaultBuffer : new byte[desiredSize];
	}

	/**
	 * Wenn der Originaldatensatz keine Daten enthält, liefert diese Methode die Daten, die stattdessen im Container (zur Markierung) gespeichert werden
	 * sollen.
	 *
	 * @param rd Datensatz
	 * @return Daten bzw. null, falls ein Fehler vorliegt
	 */
	private static byte[] writeDataModifier(ResultData rd) {
		if(rd.isNoSourceAvailable()) {
			return NO_SOURCE;
		}
		else if(rd.isNoDataAvailable()) {
			return NO_DATA;
		}
		else if(rd.isNoRightsAvailable()) {
			return NO_RIGHTS;
		}
		else {
			throw new IllegalArgumentException("Unbekannter Zustand: " + rd.getDataState());
		}
	}

	/**
	 * Serialisiert den übergebenen Datensatz und speichert ihn in {@link #actualWriteBuf} und {@link #totalWriteDataSize}. Da der
	 * ArchivRequestedTask vor dem Archivieren wissen muss, ob der Container zu lang wird und abgeschlossen werden muss, ist das
	 * Anhängen eines Datensatzes auf die Methoden serializeReqData und {@link ContainerFile#appendSerializedData(ArchiveDataSerializer)}
	 * aufgeteilt. Wenn der Datensatz ein Spezialfall ist (keine Quelle, keine Daten), wird ein speziell codierter Datensatz verwendet.
	 * <p>Diese Methode wird zum serialisieren von nachgeforderten Datensätzen verwendet.
	 *
	 * @param ad         Zu archivierender Datensatz
	 * @param arcTime    Archivzeitstempel
	 * @param compressor Deflater zum Komprimieren
	 * @return Länge des serialisierten Datensatzes
	 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
	 */
	public int serializeReqData(ArchiveData ad, long arcTime, Deflater compressor) throws PersistenceException {
		createWriteBuffers();    // Puffer (nur für schreibende Tasks) lazy anlegen

		byte[] data;
		DataState dstate = ad.getDataType();

		// Nur die drei Fälle werden hier behandelt. Lücken werden außerhalb erkannt (und
		// beim nächsten Archivsystem angefragt).

		if(dstate.equals(DataState.DATA)) {
			bos.reset();
			try {
				serializer.writeData(ad.getData());
			}
			catch(IOException e) {
				throw new PersistenceException("Datensatz konnte nicht serialisiert werden: " + e.getMessage() + "\n" + Util.ad2Str(ad), e);
			}
			data = bos.toByteArray();
		}
		else if(dstate.equals(DataState.NO_DATA)) {
			data = NO_DATA;
		}
		else if(dstate.equals(DataState.NO_SOURCE)) {
			data = NO_SOURCE;
		}
		else {
			throw new PersistenceException("DataState eines nachgeforderten Datensatzes nicht DATA/NO_DATA/NO_SOURCE: " + dstate);
		}

		actualWriteBuf = getBuf(writeBuf, getTotalDataOverhead() + data.length);    // Nach Möglichkeit bereits angelegten Puffer verwenden
		return embedSerialializedData(arcTime, ad.getDataTime(), ad.getDataIndex(), data, actualWriteBuf, compressor);
	}

	/**
	 * Serialisiert einen Datensatz der eine Datenlücke kennzeichnet und speichert ihn in {@link #actualWriteBuf} und
	 * {@link #totalWriteDataSize}. Da der ArchivTask vor dem Archivieren wissen muss, ob der Container zu lang wird
	 * und abgeschlossen werden muss, ist das Anhängen eines Datensatzes auf die Methoden
	 * {@link #serializeData(ResultData, long, Deflater)} und {@link ContainerFile#appendSerializedData(ArchiveDataSerializer)}
	 * aufgeteilt.
	 *
	 * @param dataIndex   Datenindex
	 * @param dataTime    Datenzeit
	 * @param archiveTime Archivzeit
	 * @param compressor  Komprimierer (Deflater)
	 * @return Länge des serialisierten Datensatzes
	 */
	public int serializeDataGap(long dataIndex, long dataTime, long archiveTime, Deflater compressor) {
		createWriteBuffers();    // Puffer (nur für schreibende Tasks) lazy anlegen
		actualWriteBuf = getBuf(writeBuf, getTotalDataOverhead() + POT_GAP.length);    // Nach Möglichkeit bereits angelegten Puffer verwenden
		return embedSerialializedData(archiveTime, dataTime, dataIndex, POT_GAP, actualWriteBuf, compressor);
	}

	/**
	 * Schreibt das Bytefeld mit allen Datensatz-Headern, so dass es mit {@link ContainerFile#appendSerializedData(ArchiveDataSerializer)} geschrieben
	 * werden kann. Von außen wird
	 * diese Methode nur für Testzwecke verwendet. Normalerweise sind die Methoden {@link #serializeData(ResultData, long, Deflater)} bzw.
	 * {@link #serializeReqData(ArchiveData, long, Deflater)} zu verwenden. Die Daten werden komprimiert abgelegt, wenn sie länger als
	 * {@link ContainerFile#MAX_UNCOMPRESSED} sind
	 * und die Kompression eine Verkuerzung ergeben hat.
	 *
	 * @param arcTime    Archivzeit
	 * @param dataTime   Datenzeit
	 * @param dataIdx    Datenindex
	 * @param serData    zu schreibendes Bytefeld
	 * @param dest       Ziel
	 * @param compressor Komprimierer (Deflater)
	 * @return Länge des serialisierten Datensatzes
	 */
	public int embedSerialializedData(long arcTime, long dataTime, long dataIdx, byte[] serData, byte[] dest, Deflater compressor) {
		int p = ByteIO.INT4B_LEN;                                    // [1. Platzhalter f. Länge]
		p += ByteIO.writeUnsignedLong6Byte(dest, p, arcTime);        // 2. Archivzeit
		p += ByteIO.writeUnsignedLong6Byte(dest, p, dataTime);    // 3. Datenzeit
		p += ByteIO.writeSignedLong8Byte(dest, p, dataIdx);        // 4. Datenindex
		int chkSumPos = p;                                            // [5. Platzhalter für Prüfsumme]
		p += ByteIO.INT4B_LEN;

		int netDataLen = 0;
		boolean uncompressed = true;

		if(serData.length > MAX_UNCOMPRESSED) {
			int compressInfoPos = p;                // [6. Platzhalter f. Kompressions-Info]
			p += ByteIO.INT4B_LEN;

			// Nur schreibende Tasks landen hier. Alle anderen lesen nur und brauchen keinen Deflater.
			// Notwendig, weil das Teil extrem Speicher braucht.
			if(compressor == null) compressor = new Deflater();

			compressor.reset();
			compressor.setInput(serData);
			compressor.finish();
			netDataLen = compressor.deflate(dest, p, dest.length - p - ByteIO.SEPARATOR.length);// 7. Daten

			// Hat das Komprimieren was gebracht? Wenn nicht, unkomprimierte Daten drüberschreiben
			if(netDataLen < serData.length && compressor.finished()) {
				ByteIO.writeSignedInt4Bytes(dest, compressInfoPos, serData.length);   // 6. Kompressions-Info (orig. Datenlänge)
				p += netDataLen;
				uncompressed = false;
			}
			else {
				p = compressInfoPos;
			}
		}

		if(uncompressed) {
			p += ByteIO.writeSignedInt4Bytes(dest, p, NOT_COMPRESSED);    // 6. Kompressions-Info (=0)
			p += ByteIO.writeBytes(dest, p, serData);         // 7. Daten
			netDataLen = serData.length;
		}
		totalWriteDataSize = getTotalDataOverhead() + netDataLen;     // tatsaechliche Länge über alles
		int lengthBytes = totalWriteDataSize - DATALEN_LEN;
		int checkSum = (int) (lengthBytes + arcTime + dataTime + dataIdx);

		ByteIO.writeSignedInt4Bytes(dest, 0, lengthBytes);           // 1. Länge
		ByteIO.writeSignedInt4Bytes(dest, chkSumPos, checkSum);   // 5. Prüfsumme
		ByteIO.writeSeparator(dest, p);                        // 8. Separator
		return totalWriteDataSize;
	}

	/**
	 * Gibt den Schreibpuffer zurück
	 *
	 * @return den Schreibpuffer
	 */
	public byte[] getActualWriteBuf() {
		return actualWriteBuf;
	}

	/**
	 * Gibt die benutzte Länge des Schreibpuffers zurück
	 *
	 * @return benutzte Länge in Bytes
	 */
	public int getTotalWriteDataSize() {
		return totalWriteDataSize;
	}
}
