/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import de.bsvrz.ars.ars.backup.BackupImplementation;
import de.bsvrz.ars.ars.backup.MediumNotAccessibleException;
import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.datatree.synchronization.SyncKey;
import de.bsvrz.ars.ars.persistence.BasicContainerFileHandle;
import de.bsvrz.ars.ars.persistence.ContainerFileHandle;
import de.bsvrz.ars.ars.persistence.ContainerHeaders;
import de.bsvrz.ars.ars.persistence.ContainerManagementData;
import de.bsvrz.ars.ars.persistence.ContainerManagementInformation;
import de.bsvrz.ars.ars.persistence.DeletedContainerFile;
import de.bsvrz.ars.ars.persistence.IdContainerFileDir;
import de.bsvrz.ars.ars.persistence.IdDataIdentification;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.sys.funclib.losb.util.Util;
import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;
import de.bsvrz.sys.funclib.operatingMessage.MessageState;
import de.bsvrz.sys.funclib.operatingMessage.MessageType;
import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.BlockingQueue;

/**
 * Task zum Abgleich von Verwaltungsinformationen in den Container-Headern.
 * <p>
 * Dieses Modul dient dem Abgleich von Verwaltungsinformation (=Container-Header) zwischen den Containern eines Mediums des Typs B und den entsprechenden
 * Containern im Persistenz-Verzeichnis des Archivsystems.<br> <br>
 * <p>
 * Bei jedem Aufruf werden die Header aller Container des Archivsystems durch die entsprechenden Container-Header, die auf dem spezifizierten Medium vorhanden
 * sind, ersetzt.
 *
 * @author beck et al. projects GmbH
 * @author Christian Wied
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class HeaderRestoreTask extends SingleClientSenderTask {

	private static final String TMPDIR = System.getProperty("java.io.tmpdir");

	public HeaderRestoreTask(ArchiveManager archiveMgr,BlockingQueue<Object> inputData) {
		super(archiveMgr, inputData);
	}

	@Override
	protected void work(ResultData resultData) {
		if(!resultData.hasData() || shouldTerminate()) return;
		try {
			_debug.info("Starte Abgleich der Verwaltungsinformationen");
			init(resultData.getData());
			restoreHeader();
			_debug.info("Abgleich der Verwaltungsinformationen beendet");
			sendSuccessResponse();
		}
		catch(MediumNotAccessibleException e) {
			_debug.warning("Abgleich der Verwaltungsinformationen fehlgeschlagen: Medium nicht verfügbar." + e.getMessage());
			MessageSender.getInstance().sendMessage(
					"", MessageType.APPLICATION_DOMAIN, "", MessageGrade.WARNING, MessageState.MESSAGE, "Bitte legen Sie Medium ein"
			);
			sendErrorResponse("Abgleich der Verwaltungsinformationen fehlgeschlagen: Medium nicht verfügbar." + e.getMessage());
		}
		catch(Exception e) {
			_debug.warning("Abgleich der Verwaltungsinformationen fehlgeschlagen." + e.getMessage());
			sendErrorResponse("Abgleich der Verwaltungsinformationen fehlgeschlagen: " + e.getMessage());
		}
	}


	void restoreHeader() throws Exception {
		BackupImplementation backup = BackupTask.getBackupModul();

		int mediumID = deserializer.readInt();

		List<String> contFileNames = backup.getAllContFileNames(this, mediumID);

		for(String contFile : contFileNames) {

			suspendTaskIfNecessary();	//Task anhalten, falls Archivsystem in kritischem Bereich
			if(shouldTerminate()) throw new IllegalStateException("Archivsystem wurde beendet. Verwaltungsinformationsabgleich abgebrochen");

			_debug.finer("Checking file " + contFile + " ...");
			File tmpContFile = new File(TMPDIR, contFile);
			InputStream is = backup.restoreContainer(mediumID, contFile);
			FileOutputStream fos = new FileOutputStream(tmpContFile);
			Util.copyStreams(is, fos);
			is.close();
			fos.close();
			try {
				// temporäres File auslesen, um DID + ContainerID zu bekommen:
				try(BasicContainerFileHandle tmpFileHandle = persMgr.accessContainerUnsynchronized(tmpContFile)) {
					tmpFileHandle.readContainerHeader();
					ContainerHeaders containerHeaders = new ContainerHeaders(tmpFileHandle);

					final long contId = containerHeaders.getContainerHeaderParamAsLong(ContainerManagementInformation.CHP_CONT_ID);

					// Prüfen, ob Zugriff auf Typ-A-Container erlaubt ist. Synchronisierung über DataIdentNode:
					IdDataIdentification location = containerHeaders.getLocation().getDataIdentification();
					ArchiveDataKind datakind = containerHeaders.getLocation().getArchiveDataKind();
					try(SyncKey<IdDataIdentification> indexLock = archMgr.lockIndex(location)) {
						// auf Container zugreifen
						try (ContainerFileHandle handle = persMgr.accessContainer(new IdContainerFileDir(indexLock.getElement(), datakind),
						                                                          contId)) {

							/*
							 * Aus der Spec: Beim Abgleich kann es drei Situationen geben, die wie folgt behandelt werden: <br>
							 * <ul>
							 * <li>Der Container ist im Archivsystem nicht vorhanden: In diesem Fall wird der Container-Header in
							 * das Archivsystem kopiert und dort als gelöschter und gesicherter Container markiert. Falls notwendig,
							 * wird ein entsprechendes Verzeichnis angelegt.</li>
							 * <li>der Container ist im Archivsystem vorhanden, allerdings als gelöscht markiert (d.h. es ist nur
							 * der Container-Header vorhanden). In diesem Fall wird dieser Container-Header durch den
							 * Container-Header vom Typ-B-Medium ersetzt und als gelöschter und gesicherter Container markiert. (Ein
							 * Test auf Gleichheit ist unnötig).</li>
							 * <li>Der Container ist im Archivsystem ungelöscht vorhanden. In diesem Fall werden die Header
							 * verglichen: Sind die Header vollständig identisch (modulo Reihenfolge der Schlüsselwörter), passiert
							 * nichts weiter. Sind sie nicht identisch, wird der Container gelöscht und der Container-Header durch
							 * den Container-Header vom Typ-B-Medium ersetzt und als gelöschter und gesicherter Container markiert.</li>
							 * </ul>
							 */

							// Fallunterscheidung:
							if (!handle.existsContainer()) {
								// 1. Container ist im Archiv nicht vorhanden
								// --> Header speichern und als geloescht+gesichert markieren

								// 2. Container ist vorhanden, aber als geloescht markiert
								// --> Header ersetzen und als geloescht+gesichert markieren

								DeletedContainerFile file = persMgr.openDeletedContainerFile(indexLock, datakind);
								file.addDeletedContainer(containerHeaders, tmpFileHandle);
								file.write();
							} else {
								// 3. Container ist ungeloescht vorhanden
								// --> Header vergleichen
								handle.readContainerHeader();
								if (!isHeaderIdentical(handle, containerHeaders)) {
									DeletedContainerFile file = persMgr.openDeletedContainerFile(indexLock, datakind);
									file.addDeletedContainer(containerHeaders, tmpFileHandle);
									file.write();

									File originalFile = handle.getFile();
									handle.close();
									if (!originalFile.delete()) {
										throw new PersistenceException("Kann Containerdatei nicht löschen: " + originalFile);
									}
								} else {
									handle.close();
								}
							}
						}
					}
				}
			}
			finally {
				tmpContFile.delete();
			}
		} // end for
	}

	/**
	 * Prüft, ob der Header des momentan im Zugriff des PersistenceManager befindlichen Containers mit dem des übergebenen Files identisch ist. Die Reihenfolge
	 * der Header-Parameter darf unterschiedlich sein.
	 *
	 *
	 * @param handle
	 * @param other
	 *
	 * @return true, wenn Header identisch sind
	 */
	private static boolean isHeaderIdentical(final ContainerManagementData handle, ContainerManagementData other) {

		try {
			Map<ContainerManagementInformation, String> paramValues1 = new EnumMap<>(ContainerManagementInformation.class);
			for(ContainerManagementInformation param : ContainerManagementInformation.values()) {
				paramValues1.put(param, handle.getContainerHeaderParamAsString(param));
			}
			Map<ContainerManagementInformation, String> paramValues2 = new EnumMap<>(ContainerManagementInformation.class);
			for(ContainerManagementInformation param : ContainerManagementInformation.values()) {
				paramValues2.put(param, other.getContainerHeaderParamAsString(param));
			}
			return paramValues1.equals(paramValues2);
		}
		catch(Exception e) {
			_debug.warning("Fehler beim Abgleich der Container-Header: " + other, e);
		}
		return false;
	}
}
