/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.sys.funclib.kappich.annotations.NotNull;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.Executors;
import java.util.concurrent.SynchronousQueue;

/**
 * Abstrakte Basisklasse für die abstrakten Task-Basisklassen {@link de.bsvrz.ars.ars.mgmt.tasks.SingleTask} und {@link de.bsvrz.ars.ars.mgmt.tasks.MultiTask}.
 * <p>
 * Ein QueueTask ist ein Thread, der laufend eine bestimmte Aktion ausführt. Er ähnelt einem {@link Executors#newSingleThreadExecutor()}, die einzelnen Aufgaben
 * werden aber nicht als Runnable o.ä. übergeben, sondern der Task holt sich selbst die durchzuführenden Arbeitsschritte aus einer Queue. Die Klasse, auf der
 * ein Task arbeitet wird über den generischen Typ T definiert.
 *
 * @author beck et al. projects GmbH
 * @author Thomas Schaefer
 * @author Alexander Schmidt
 * @author Kappich Systemberatung
 * @version $Revision$ / $Date$ / ($Author$)
 */
public abstract class QueueTask<T> extends RepeatingTask {

	private final BlockingQueue<Object> _queue;

	protected QueueTask(ArchiveManager archiveMgr, BlockingQueue<Object> queue) {
		super(archiveMgr);
		_queue = queue;
		if(queue instanceof SynchronousQueue) {
			/*
			Es ist erforderlich, dass immer mindestens 1 Element in der Queue Platz hat um das terminieren Deadlockfrei zu gestalten.
			Sonst kann es passieren, dass ein Auftrag abgearbeitet wird, in #execute bereits gerpüft wurde, ob der Thread angehalten wurde (nein),
			dann #terminateTask aufgerufen wird, was _shouldTerminate (zu spät) auf true setzt und ebenfalls das Terminierungstoken nicht einfügen kann,
			weil noch nicht queue.take ausgeführt wird. 
			 */
			throw new IllegalArgumentException("SynchronousQueue darf nicht benutzt werden, Deadlockgefahr");
		}
	}

	/**
	 * Fügt einen abzuarbeitenden Auftrag in die Queue ein.
	 * @param step Datensatz
	 * @throws InterruptedException Thread unterbrochen
	 */
	public void submit(@NotNull T step) throws InterruptedException {
		_queue.put(step);
	}
	
	/**
	 * Fügt einen abzuarbeitenden Auftrag in die Queue ein, wenn das möglich ist.
	 * @param step Datensatz
	 * @throws InterruptedException Thread unterbrochen
	 * @return true wenn eingefügt wurde, sonst false
	 */
	public boolean offer(@NotNull T step) throws InterruptedException {
		return _queue.offer(step);
	}

	/**
	 * Terminiert den Task. Diese Methode fügt zusätzlich in die Queue einen speziellen Token ein, mit dem der Aufruf von
	 * {@code _queue.take()} in der {@link #step()}-Methode aufgeweckt wird, falls die Methode dort gerade auf einen neuen Auftrag warten.
	 */
	@Override
	public void terminateTask() {
		super.terminateTask();
		if(_queue != null) _queue.offer(TERMINATE_TASK);
	}

	@Override
	public final void step() throws InterruptedException {
		Object take = _queue.take();
		//noinspection ObjectEquality
		if(take == TERMINATE_TASK) {
			return;
		}

		// Die Queue enthält entweder TERMINATE_TASK oder Objekte vom Typ T
		work((T) take);
		taskStepDone();
	}

	/**
	 * Vom konkreten Task zu implementierende Methode, die den Auftrag abarbeitet
	 * @param step Auftrag/Arbeitsschritt
	 * @throws InterruptedException Thread unterbrochen
	 */
	protected abstract void work(T step) throws InterruptedException;

}
