/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009-2019 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence;

import com.google.common.collect.ImmutableMap;
import de.bsvrz.ars.ars.persistence.index.*;
import de.bsvrz.ars.ars.persistence.index.backend.management.LongRange;
import de.bsvrz.ars.ars.persistence.index.result.IndexResult;
import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Mit den statischen Methoden dieser Klasse können die Container in einem Verzeichnis geprüft werden.
 * <p>
 * simpleCheck prüft dabei nur die Container-Header, extendedCheck überprüft auch die Nutzdaten.
 *
 * @author beck et al. projects GmbH
 * @author Phil Schrettenbrunner
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ConsistencyCheck {

	public static final int OK = 0;

	public static final int WARN = 1;

	public static final int ERROR = 2;

	private static final Pattern SimVarPattern = Pattern.compile(".*sv(\\d{1,3}).*");

	/** Array Index für MonotoneIndex */
	private static final int MON_DIMIN = 0;

	private static final int MON_DIMAX = 1;

	private static final int MON_ATMIN = 2;

	private static final int MON_ATMAX = 3;

	private static final int MON_DTMIN = 4;

	private static final int MON_DTMAX = 5;

	/** Array Index für DataTimeIndex */
	private static final int DTI_DTMIN = 0;

	private static final int DTI_DTMAX = 1;

	private static final int DTI_DIMIN = 2;

	private static final int DTI_DIMAX = 3;

	/** Array Index für DataIndexIndex */
	private static final int DII_DIMIN = 0;

	private static final int DII_DIMAX = 1;

	/**
	 * Prüft alle Container-Header im angegebenen Verzeichnis.
	 * <p>
	 * Dabei wird geprüft <ul> <li>ob das Verzeichnis existiert und beschreibbar ist</li> <li>ob die IndexDateien existieren</li> <li>ob das Verzeichnis
	 * Container-Dateien enthält</li> <li>ob der Container abgeschlossen ist</li> <li>ob für jeden Container die containterID im Header == containerID im
	 * Dateinamen ist</li> <li>ob für jeden Container der DatenIndexIndex-Min Wert &lt; DatenIndexIndex-Max Wert im Header ist</li> <li>ob für jeden Container
	 * der DatenZeitIndex-Min Wert &lt;= DatenZeitIndex-Max Wert im Header ist</li> <li>ob für jeden Container der ArchivZeitIndex-Min Wert &lt;=
	 * ArchivZeitIndex-Max Wert im Header ist</li> <li>ob es IndexEinträge für nichtexistierende Container gibt</li> <li>ob es Container gibt, für die es keine
	 * IndexEinträge gibt</li> </ul>
	 *
	 * @param dir    Das Verzeichnis, in dem die Container-Header geprüft werden.
	 * @param logger Logger, auf dem Meldungen ausgegeben werden sollen
	 *
	 * @return Binärwert, der angibt, ob alles OK war, ob Warnungen oder Fehler aufgetreten sind. (OK, WARN, ERROR und WARN/ERROR)
	 */
	public static int simpleCheck(File dir, Debug logger) {
		return check(dir, logger, false);
	}


	/**
	 * Prüft alle Container (Header und Body) im angegebenen Verzeichnis.
	 * <p>
	 * Dabei wird geprüft <ul> <li>ob das Verzeichnis existiert und beschreibbar ist</li> <li>ob die IndexDateien existieren</li> <li>ob das Verzeichnis
	 * Container-Dateien enthält</li> <li>ob der Container abgeschlossen ist</li> <li>ob für jeden Container die containterID im Header == containerID im
	 * Dateinamen ist</li> <li>ob für jeden Container der DatenIndexIndex-Min Wert &lt; DatenIndexIndex-Max Wert im Header ist</li> <li>ob für jeden Container
	 * der DatenZeitIndex-Min Wert &lt;= DatenZeitIndex-Max Wert im Header ist</li> <li>ob für jeden Container der ArchivZeitIndex-Min Wert &lt;=
	 * ArchivZeitIndex-Max Wert im Header ist</li> <li>ob für jeden Eintrag in jedem Container der DatenIndexIndex-Min<sub>n-1</sub> Wert &lt;
	 * DatenIndexIndex-Min<sub>n</sub> ist</li> <li>ob für jeden Eintrag in jedem Container der DatenZeitIndex-Min<sub>n-1</sub> Wert &lt; DatenZeitIndex Min
	 * <sub>n</sub> ist</li> <li>ob für jeden Eintrag in jedem Container der ArchivZeitIndex-Max<sub>n-1</sub> Wert &lt; ArchivZeitIndex-Max<sub>n</sub>
	 * ist</li> <li>ob für jeden Eintrag in jedem Container der DatenIndexIndex-Max<sub>n-1</sub> Wert &lt; DatenIndexIndex-Max<sub>n</sub> ist</li> <li>ob für
	 * jeden Eintrag in jedem Container der DatenZeitIndex-Max<sub>n-1</sub> Wert &lt; DatenZeitIndex-Max<sub>n</sub> ist</li> <li>ob für jeden Eintrag in jedem
	 * Container der ArchivZeitIndex-Max<sub>n-1</sub> Wert &lt; ArchivZeitIndex-Max<sub>n</sub> ist</li> <li>ob für jeden Container der Headereintrag
	 * Anzahl-Datensätze mit der tatsächlichen Anzahl der Datensaätze übereinstimmt.</li> <li>ob es IndexEinträge für nichtexistierende Container gibt</li>
	 * <li>ob es Container gibt, für die es keine IndexEinträge gibt</li> </ul>
	 *
	 * @param dir    Das Verzeichnis, in dem die Container geprüft werden.
	 * @param logger Logger, auf dem Meldungen ausgegeben werden sollen oder {@code null} falls keine Meldungen ausgegeben werden sollen.
	 *
	 * @return Binärwert, der angibt, ob alles OK war, ob Warnungen oder Fehler aufgetreten sind. (OK, WARN, ERROR und WARN/ERROR)
	 */
	public static int extendedCheck(File dir, Debug logger) {
		return check(dir, logger, true);
	}


	private static int check(File dir, Debug logger, boolean doExtendedCheck) {
		int status = OK;

		// //////////////////////////////////////////////////////////////////
		// Check auf den Ordner
		if(!dir.exists()) {
			logFiner(logger, dir.toString() + ": Verzeichnis existiert nicht.");
			return ERROR;
		}
		if(!dir.canRead()) {
			logFiner(logger, dir.toString() + ": Verzeichnis kann nicht gelesen werden.");
			return ERROR;
		}
		if(!dir.isDirectory()) {
			logFiner(logger, dir.toString() + ": Ist kein Verzeichnis.");
			return ERROR;
		}

		// SimVar aus Verzeichnisnamen extrahieren
		Matcher m = SimVarPattern.matcher(dir.toString());
		int SimVar = -1;
		if(m.matches()) SimVar = Integer.parseInt(m.group(1));

		// //////////////////////////////////////////////////////////////////
		// Indexe laden
		HashMap<Long, long[]> monoIdx = null, dtIndex = null, dIdxIndex = null;
		try {
			dtIndex = loadDataTimeIdx(new File(dir, DataTimeIndexImpl.IDX_FILENAME));
		}
		catch(IOException e1) {
			logFinest(logger, dir.toString() + ": DataTimeIndex: Keine Indexdatei.");
			status |= WARN;
		}
		try {
			monoIdx = loadMonotoneIndex(new File(dir, ArchiveTimeIndexImpl.IDX_FILENAME));
		}
		catch(IOException e1) {
			logFinest(logger, dir.toString() + ": MonotoneIndex: Keine Indexdatei.");
			status |= WARN;
		}

		ArchiveDataKind adk = PersistenceManager.getArchiveDataKindFromDir(dir);
		if(adk.isRequested()) {
			try {
				dIdxIndex = loadDataIndexeIdx(new File(dir, DataIndexIndexImpl.IDX_FILENAME));
			}
			catch(IOException e1) {
				logFinest(logger, dir.toString() + ": DatenindexIndex: Keine Indexdatei.");
				status |= WARN;
			}
		}

		ContainerManagementIndex managementIndex = null;
		try {
			try {
				managementIndex = new ContainerManagementIndex(1024, Paths.get(dir.getPath()).resolve(ContainerManagementIndex.IDX_FILENAME));
			}
			catch(IndexException e1) {
				logFinest(logger, dir.toString() + ": Verwaltungsdatenindex: Konnte nicht gelesen werden.");
				status |= WARN;
			}

			long ObjIdInPathname = extractId("obj", dir);
			long AtgIdInPathname = extractId("atg", dir);
			long AspIdInPathname = extractId("asp", dir);

			File[] cont = dir.listFiles(ContainerFile.CONT_FILENAME_FILTER);

			if(cont == null || cont.length == 0) {
				logFiner(logger, dir.toString() + ": Verzeichnis enthält keine Container-Dateien.");
				return ERROR;
			}
			Arrays.sort(cont);

			// //////////////////////////////////////////////////////////////////
			// Alle Container abarbeiten

			long lastDataIndex = Long.MIN_VALUE, lastArcTime = Long.MIN_VALUE, lastDataTime = Long.MIN_VALUE;

			int totalDS = 0;
			for(int j = 0; j < cont.length; j++) {
				File file = cont[j];
				try (BasicContainerFileHandle cf = new BasicContainerFileHandle(file)) {
					cf.readContainerHeader();

managementIf:
					if (managementIndex != null) {
						IndexResult<ContainerManagementInformation> indexResult = null;
						try {
							indexResult = managementIndex.query(ImmutableMap.of(ContainerManagementInformation.CHP_CONT_ID, new LongRange(cf.getContainerId(), cf.getContainerId())));
						} catch (IndexException ignored) {
							logFiner(logger, managementIndex.getFile() + ": Indexdatei konnte nicht verarbeitet werden");
							status |= WARN;
						}

						if (indexResult == null || indexResult.isEmpty() || indexResult.get(0, ContainerManagementInformation.CHP_CONT_ID) != cf.getContainerId()) {
							logFiner(logger, managementIndex.getFile() + ": Datei enthält den Container " + cf.getContainerId() + " nicht.");
							status |= WARN;
							break managementIf;
						}

						// Vergleichen der ContainerHeader mit dem Eintrag im Index
						for(ContainerManagementInformation value : ContainerManagementInformation.values()) {
							if(value.isNumeric()) {
								if(value.getByteLength() <= 4) {
									if(cf.getContainerHeaderParamAsInt(value.delegate()) != indexResult.getSignedInt(0, value)) {
										logFiner(logger, managementIndex.getFile() + ": Index enthält falschen Wert für Container " + cf.getContainerId() + " und Information " + value + "(" + cf
												.getContainerHeaderParamAsInt(value.delegate()) + "!=" + indexResult.getSignedInt(0, value) + ")");
										status |= ERROR;
									}
								}
								else {
									if(cf.getContainerHeaderParamAsLong(value.delegate()) != indexResult.get(0, value)) {
										logFiner(logger, managementIndex.getFile() + ": Index enthält falschen Wert für Container " + cf.getContainerId() + " und Information " + value + "(" + cf
												.getContainerHeaderParamAsLong(value.delegate()) + "!=" + indexResult.get(0, value) + ")");
										status |= ERROR;
									}
								}
							}
							else {
								if(!Objects.equals(cf.getContainerHeaderParamAsString(value.delegate()), indexResult.getString(0, value))) {
									logFiner(logger, managementIndex.getFile() + ": Index enthält falschen Wert für Container " + cf.getContainerId() + " und Information " + value + "(" + cf
											.getContainerHeaderParamAsString(value.delegate()) + "!=" + indexResult.getString(0, value) + ")");
									status |= ERROR;
								}
							}
						}
					}

					// Vergleichen der Datenspezifikation in Header und Pfadnamen
					if (cf.getContainerHeaderParamAsLong(ContainerHdr.CHP_OBJ_ID) != ObjIdInPathname) {
						logFiner(
								logger, file.getName() + ": Container-Header: ObjektID im Header (" + cf.getContainerHeaderParamAsLong(ContainerHdr.CHP_OBJ_ID)
										+ ") entspricht nicht der ObjektId (" + ObjIdInPathname + ") im Pfad."
						);
						status |= ERROR;
					}
					if (cf.getContainerHeaderParamAsLong(ContainerHdr.CHP_ATG_ID) != AtgIdInPathname) {
						logFiner(
								logger, file.getName() + ": Container-Header: Attributgruppe im Header (" + cf.getContainerHeaderParamAsLong(ContainerHdr.CHP_ATG_ID)
										+ ") entspricht nicht der Atributgruppe (" + AtgIdInPathname + ") im Pfad."
						);
						status |= ERROR;
					}
					if (cf.getContainerHeaderParamAsLong(ContainerHdr.CHP_ASP_ID) != AspIdInPathname) {
						logFiner(
								logger, file.getName() + ": Container-Header: Aspekt im Header (" + cf.getContainerHeaderParamAsLong(ContainerHdr.CHP_ASP_ID)
										+ ") entspricht nicht dem Aspekt (" + AspIdInPathname + ") im Pfad."
						);
						status |= ERROR;
					}

					long cfContIDFromFileName = PersistenceManager.getContID(file);

					long[] monoIdxEntry = null;
					if (monoIdx != null) {
						monoIdxEntry = monoIdx.remove(cfContIDFromFileName);
						if (monoIdxEntry == null) {
							logFiner(logger, file.getAbsolutePath() + ": MonotoneIndex: Kein Eintrag im Index für diesen Container");
							status |= WARN;
						}
					}

					long[] dtIdxEntry = null;
					if (dtIndex != null) {
						dtIdxEntry = dtIndex.remove(cfContIDFromFileName);
						if (dtIdxEntry == null) {
							logFiner(logger, file.getAbsolutePath() + ": DataTime: Kein Eintrag im Index für diesen Container");
							status |= WARN;
						}
					}

					long[] dIdxIdxEntry = null;
					if (dIdxIndex != null) {
						dIdxIdxEntry = dIdxIndex.remove(cfContIDFromFileName);
						if (dIdxIdxEntry == null) {
							logFiner(logger, file.getAbsolutePath() + ": DataIndex: Kein Eintrag im Index für diesen Container");
							status |= WARN;
						}
					}

					long dIdxMin = cf.getContainerHeaderParamAsLong(ContainerHdr.CHP_DATA_IDX_MIN);
					long dIdxMax = cf.getContainerHeaderParamAsLong(ContainerHdr.CHP_DATA_IDX_MAX);
					long aTMin = cf.getContainerHeaderParamAsLong(ContainerHdr.CHP_ARC_TIME_MIN);
					long aTMax = cf.getContainerHeaderParamAsLong(ContainerHdr.CHP_ARC_TIME_MAX);
					long dTMin = cf.getContainerHeaderParamAsLong(ContainerHdr.CHP_DATA_TIME_MIN);
					long dTMax = cf.getContainerHeaderParamAsLong(ContainerHdr.CHP_DATA_TIME_MAX);
					long anzDS = cf.getContainerHeaderParamAsLong(ContainerHdr.CHP_ANZ_DS);

					// Prüfen der Abschlusskriterien: Nicht abgeschlossene Contaiener haben 0 bzw. -1 in diesen
					// Parametern stehen, abgeschlossene Container die entsprechenden Werte. Es müssen entweder alle
					// oder keiner dieser Parameter Abschlusswerte haben.
					if ((dIdxMin != 0 || dIdxMax != 0 || aTMin != 0 || aTMax != 0 || dTMin != 0 || dTMax != 0 || anzDS != -1) && !(dIdxMin != 0 && dIdxMax != 0
							&& aTMin != 0 && aTMax != 0
							&& dTMin != 0 && dTMax != 0
							&& anzDS != -1)) {
						logFiner(logger, file.getName() + ": Container-Header: Abschlusskriterien nicht einheitlich");
						status |= ERROR;
					}

					if (cf.isContainerClosed()) {
						// //////////////////////////////////////////////////////
						// Einfacher check der Header Information

						if (dIdxMin > dIdxMax) {
							logFiner(logger, file.getName() + ": Container-Header: DataIndexIndex Min > Max");
							status |= ERROR;
						}
						if (aTMin > aTMax) {
							logFiner(logger, file.getName() + ": Container-Header: ArchiveTimeIndex Min > Max");
							status |= ERROR;
						}
						if (dTMin > dTMax) {
							logFiner(logger, file.getName() + ": Container-Header: DataTime Min > Max");
							status |= ERROR;
						}

						// Check, dass die im Header stehende Container Id zum Dateinamen passt
						long cfContId = cf.getContainerHeaderParamAsLong(ContainerHdr.CHP_CONT_ID);
						if (!String.format("dc%013d.dat", cfContId).equals(file.getName())) {
							logFiner(logger, file.getAbsolutePath() + ": Container-Header: ContainerId im Header != ContainerId im Dateinamen");
							status |= ERROR;
						}

						// Prüfen der Indices
						if (dtIdxEntry != null) {
							if (!(dtIdxEntry[DTI_DTMIN] == dTMin && dtIdxEntry[DTI_DTMAX] == dTMax && dtIdxEntry[DTI_DIMIN] == dIdxMin
									&& dtIdxEntry[DTI_DIMAX] == dIdxMax)) {
								logFiner(logger, file.getAbsolutePath() + ": DataTimeIndex: Indexeintrag stimmt nicht mit Containerheader überein");
								status |= WARN;
							}
						}
						if (monoIdxEntry != null) {
							if (!(monoIdxEntry[MON_DIMIN] == dIdxMin && monoIdxEntry[MON_DIMAX] == dIdxMax && monoIdxEntry[MON_ATMIN] == aTMin
									&& monoIdxEntry[MON_ATMAX] == aTMax && monoIdxEntry[MON_DTMIN] == dTMin && monoIdxEntry[MON_DTMAX] == dTMax)) {
								logFiner(logger, file.getAbsolutePath() + ": MonotoneIndex: Indexeintrag stimmt nicht mit Containerheader überein.");
								status |= WARN;
							}
						}
						if (dIdxIdxEntry != null) {
							if (!(dIdxIdxEntry[DII_DIMIN] == dIdxMin && dIdxIdxEntry[DII_DIMAX] == dIdxMax)) {
								logFiner(logger, file.getAbsolutePath() + ": DataindexIndex: Indexeintrag stimmt nicht mit Containerheader überein.");
								status |= WARN;
							}
						}

						// Prüfen der Container Art
						try {
							if (!dir.getName().equalsIgnoreCase(cf.getContainerHeaderParamAsString(ContainerHdr.CHP_DATA_KIND))) {
								logFine(
										logger, file.getName() + ": Datenart im Verzeichnisnamen (" + dir.getName() + ") stimmt nicht mit Headerinformation ("
												+ cf.getContainerHeaderParamAsString(ContainerHdr.CHP_DATA_KIND) + ") überein."
								);
								status |= ERROR;
							}
						} catch (PersistenceException e) {
							logFine(logger, file.getName() + ": Ungültige Datenart (" + cf.getContainerHeaderParamAsString(ContainerHdr.CHP_DATA_KIND) + ") im Header.");
							status |= ERROR;
						}

						// Prüfen der SimVar
						if (SimVar != cf.getContainerHeaderParamAsInt(ContainerHdr.CHP_SIM_VAR)) {
							logFine(
									logger, file.getName() + ": Simulationsvariante im Verzeichnisnamen (" + SimVar + ") stimmt nicht mit Headerinformation ("
											+ cf.getContainerHeaderParamAsInt(ContainerHdr.CHP_SIM_VAR) + ") überein."
							);
							status |= ERROR;
						}

						// /////////////////////////////////////////////////////////////////////////////
						// Extended checks auf die Nutzdaten eines Containers
						if (doExtendedCheck) {
							// Min/Max Werte der Nutzdaten/Paylod ermitteln
							long plDataIndexMin = Long.MAX_VALUE, plDataIndexMax = Long.MIN_VALUE;
							long plArcTimeMin = Long.MAX_VALUE, plArcTimeMax = Long.MIN_VALUE;
							long plDataTimeMin = Long.MAX_VALUE, plDataTimeMax = Long.MIN_VALUE;
							long plCount = 0;

							ContainerDataIterator i = new ContainerDataIterator(cf, null);
							ContainerDataResult result = new ContainerDataResult();
							while (!i.isEmpty()) {
								plCount++;
								i.peek(result);
								// DataIndex
								if (result.getDataIndex() > plDataIndexMax) plDataIndexMax = result.getDataIndex();
								if (result.getDataIndex() < plDataIndexMin) plDataIndexMin = result.getDataIndex();
								if (lastDataIndex >= result.getDataIndex()) {
									logFiner(
											logger, file.toString() + ": DatenIndex nicht streng monoton steigend\n" + "I_n-1=" + Util.dIdx2StrExt(lastDataIndex)
													+ " (" + lastDataIndex + "), I_n=" + Util.dIdx2StrExt(result.getDataIndex()) + " (" + result.getDataIndex() + ")"
									);
									status |= ERROR;
								}
								status |= checkHdrConsistency(result, file, logger);

								// ArcTime
								if (result.getArchiveTime() > plArcTimeMax) plArcTimeMax = result.getArchiveTime();
								if (result.getArchiveTime() < plArcTimeMin) plArcTimeMin = result.getArchiveTime();
								if (lastArcTime > result.getArchiveTime()) {
									logFiner(
											logger, file.toString() + ": Archivzeit nicht monoton steigend\n" + "I_n-1=" + Util.timestrMillisFormatted(lastArcTime)
													+ " (" + lastArcTime + "), I_n=" + Util.timestrMillisFormatted(result.getArchiveTime()) + " (" + result.getArchiveTime() + ")"
									);
									status |= ERROR;
								}
								// DataTime
								if (result.getDataTime() > plDataTimeMax) plDataTimeMax = result.getDataTime();
								if (result.getDataTime() < plDataTimeMin) plDataTimeMin = result.getDataTime();
								
								lastDataIndex = result.getDataIndex();
								lastArcTime = result.getArchiveTime();

								i.remove();
							}
							i.close();

							// Anzahl Datensätze
							if (plCount != cf.getContainerHeaderParamAsLong(ContainerHdr.CHP_ANZ_DS)) {
								logFiner(
										logger, file.toString() + ": Anzahl Datensätze: Headerinformation (" + cf.getContainerHeaderParamAsLong(ContainerHdr.CHP_ANZ_DS)
												+ ") stimmt nicht mit den Nutzdaten (" + plCount + ") überein."
								);
								status |= ERROR;
							}

							// Indexe
							if (plDataIndexMin != dIdxMin) {
								logFiner(
										logger, file.toString() + ": DataIndexMin: Headerinformation (" + dIdxMin + ") stimmt nicht mit den Nutzdaten ("
												+ plDataIndexMin + ") überein."
								);
								status |= ERROR;
							}
							if (plDataIndexMax != dIdxMax) {
								logFiner(
										logger, file.toString() + ": DataIndexMax: Headerinformation (" + dIdxMax + ") stimmt nicht mit den Nutzdaten ("
												+ plDataIndexMax + ") überein."
								);
								status |= ERROR;
							}
							if (plArcTimeMin != aTMin) {
								logFiner(
										logger, file.toString() + ": ArcTimeMin: Headerinformation (" + aTMin + ") stimmt nicht mit den Nutzdaten (" + plArcTimeMin
												+ ") überein."
								);
								status |= ERROR;
							}
							if (plArcTimeMax != aTMax) {
								logFiner(
										logger, file.toString() + ": ArcTimeMax: Headerinformation (" + aTMax + ") stimmt nicht mit den Nutzdaten (" + plArcTimeMax
												+ ") überein."
								);
								status |= ERROR;
							}
							if (plDataTimeMin != dTMin) {
								logFiner(
										logger, file.toString() + ": DataTimeMin: Headerinformation (" + dTMin + ") stimmt nicht mit den Nutzdaten ("
												+ plDataTimeMin + ") überein."
								);
								status |= ERROR;
							}
							if (plDataTimeMax != dTMax) {
								logFiner(
										logger, file.toString() + ": DataTimeMax: Headerinformation (" + dTMax + ") stimmt nicht mit den Nutzdaten ("
												+ plDataTimeMax + ") überein."
								);
								status |= ERROR;
							}
						}
					} else {
						// Container noch nicht abgeschlossen
						if (j != cont.length - 1) {
							// Container mittendrin nicht abgeschlossen -> Fehler
							logFiner(logger, file.toString() + ": Container nicht abgeschlossen und es gibt Folgecontainer.");
							status |= ERROR;
						}
					}
				} catch (PersistenceException e) {
					logError(logger, file.getName() + ": Fehler beim Zugriff: " + e.getMessage());
					status |= ERROR;
				}
				totalDS++;
			}

			// Prüfen, ob es noch Index-Einträge gibt, zu denen es keine Container (und die somit
			// noch nicht aus der Liste gelöscht worden sind) gibt.

			if(dtIndex != null && dtIndex.size() > 0) {
				logFiner(logger, dir.toString() + ": Es gibt Einträge im DataTimeIndex, zu denen es keine Container gibt. " + hashMapToString(dtIndex));
				status |= ERROR;
			}
			if(monoIdx != null && monoIdx.size() > 0) {
				logFiner(logger, dir.toString() + ": Es gibt Einträge im monotonen Index, zu denen es keine Container gibt. " + hashMapToString(monoIdx));
				status |= ERROR;
			}
		}
		finally {
			if(managementIndex != null) {
				try {
					managementIndex.close();
				}
				catch(IndexException e) {
					logError(logger, managementIndex.getFile() + ": Fehler beim Schließen des Verwaltungsdaten-Index: " + e.getMessage());
					status |= ERROR;
				}
			}
		}
		return status;
	}

	private static int checkHdrConsistency(ContainerDataResult iter, File file, Debug logger) {
		int status = OK;
		if(iter.getDataState() == DataState.NO_SOURCE && (Util.dIdxDaVBit(iter.getDataIndex()) != 1 || Util.dIdxArSBit(iter.getDataIndex()) != 0)) {
			logFiner(
				logger, "Datensatz mit Kennung 'keine Quelle' hat DAV-Bit nicht gesetzt oder Archiv-Bit gesetzt: Pos. " + iter.getDataIndex() + ", "
				        + file.getAbsolutePath()
			);
			status |= ERROR;
		}
		if(iter.getDataState() == DataState.NO_DATA && (Util.dIdxDaVBit(iter.getDataIndex()) != 0 || Util.dIdxArSBit(iter.getDataIndex()) != 0)) {
			logFiner(
				logger, "Datensatz mit Kennung 'keine Daten' hat DAV-Bit oder Archiv-Bit gesetzt: Pos. " + iter.getDataIndex() + ", " + file.getAbsolutePath()
			);
			status |= ERROR;
		}
		if(iter.isData() && (Util.dIdxDaVBit(iter.getDataIndex()) != 0 || Util.dIdxArSBit(iter.getDataIndex()) != 0)) {
			logFiner(logger, "regulaerer Datensatz hat DAV-Bit oder Archiv-Bit gesetzt: Pos. " + iter.getDataIndex() + ", " + file.getAbsolutePath());
			status |= ERROR;
		}
		if(iter.getDataState() == DataState.POSSIBLE_GAP && (Util.dIdxDaVBit(iter.getDataIndex()) != 0 || Util.dIdxArSBit(iter.getDataIndex()) != 1)) {
			logFiner(
				logger, "Datensatz mit Kennung 'Datenluecke' hat DAV-Bit gesetzt oder Archiv-Bit nicht gesetzt: Pos. " + iter.getDataIndex() + ", "
				        + file.getAbsolutePath()
			);
			status |= ERROR;
		}
		if(Util.dIdxSrcSubscrTime(iter.getDataIndex()) == 0) {
			logFiner(logger, "Datenindex mit AnmeldezeitQuelle==0 gefunden: Pos. " + iter.getDataIndex() + ", " + file.getAbsolutePath());
			status |= ERROR;
		}

		return status;
	}

	private static HashMap<Long, long[]> loadMonotoneIndex(File indexFile) throws IOException {
		byte[] bytes = new byte[(int) indexFile.length()];
		HashMap<Long, long[]> hm = new HashMap<>((int) indexFile.length() / IndexEntryMonotone.ENTRY_LEN);

		try(RandomAccessFile raf = new RandomAccessFile(indexFile, "r")) {
			raf.read(bytes, 0, bytes.length);
		}

		// Da Reihenfolge bei HashMap egal ist, können wir sie auch rückwerts füllen -> schneller
		for(int idx = (bytes.length / IndexEntryMonotone.ENTRY_LEN) - 1; idx >= 0; idx--) {
			long[] l = new long[6];
			l[MON_DIMIN] = IndexEntryMonotone.getDIdxMin(bytes, idx);
			l[MON_DIMAX] = IndexEntryMonotone.getDIdxMax(bytes, idx);
			l[MON_ATMIN] = IndexEntryMonotone.getATimeMin(bytes, idx);
			l[MON_ATMAX] = IndexEntryMonotone.getATimeMax(bytes, idx);
			l[MON_DTMIN] = IndexEntryMonotone.getDTimeMin(bytes, idx);
			l[MON_DTMAX] = IndexEntryMonotone.getDTimeMax(bytes, idx);
			hm.put(IndexEntryMonotone.getCID(bytes, idx), l);
		}
		return hm;
	}


	private static HashMap<Long, long[]> loadDataTimeIdx(File indexFile) throws IOException {
		HashMap<Long, long[]> hm = new HashMap<>();

		byte[] bytes = new byte[(int) indexFile.length()];

		try(RandomAccessFile raf = new RandomAccessFile(indexFile, "r")) {
			raf.read(bytes, 0, bytes.length);
		}

		// Da Reihenfolge bei HashMap egal ist, können wir sie auch rückwerts füllen -> schneller
		for(int idx = (bytes.length / IndexEntryTime.ENTRY_LEN) - 1; idx >= 0; idx--) {
			long[] l = new long[4];
			l[DTI_DIMIN] = IndexEntryTime.getDIMin(bytes, idx);
			l[DTI_DIMAX] = IndexEntryTime.getDIMax(bytes, idx);
			l[DTI_DTMIN] = IndexEntryTime.getDTMin(bytes, idx);
			l[DTI_DTMAX] = IndexEntryTime.getDTMax(bytes, idx);
			hm.put(IndexEntryTime.getCID(bytes, idx), l);
		}
		return hm;
	}


	private static HashMap<Long, long[]> loadDataIndexeIdx(File indexFile) throws IOException {
		HashMap<Long, long[]> hm = new HashMap<>();

		byte[] bytes = new byte[(int) indexFile.length()];

		try(RandomAccessFile raf = new RandomAccessFile(indexFile, "r")) {
			raf.read(bytes, 0, bytes.length);
		}

		// Da Reihenfolge bei HashMap egal ist, können wir sie auch rückwerts füllen -> schneller
		for(int idx = (bytes.length / IndexEntryIndex.ENTRY_LEN) - 1; idx >= 0; idx--) {
			long[] l = new long[2];
			l[DII_DIMIN] = IndexEntryIndex.getDIMin(bytes, idx);
			l[DII_DIMAX] = IndexEntryIndex.getDIMax(bytes, idx);
			hm.put(IndexEntryIndex.getCID(bytes, idx), l);
		}
		return hm;
	}


	private static String hashMapToString(HashMap<Long, long[]> hashMap) {
		if(hashMap == null || hashMap.size() == 0) return "";

		StringBuilder sb = new StringBuilder();
		sb.append("{");
		for(final Map.Entry<Long, long[]> longEntry : hashMap.entrySet()) {
			long[] data = longEntry.getValue();
			sb.append("(");
			for(final long aData : data) {
				sb.append(aData + ", ");
			}
			sb.append(longEntry.getKey() + ") ");
		}
		return sb.append("}").toString();
	}


	private static long extractId(String id, File path) {
		String[] elements = path.toString().split(File.separator.equals("\\") ? "\\\\" : File.separator);

		StringBuilder sb = new StringBuilder();

		for(String element : elements) {
			if(element.startsWith(id)) sb.append(element, id.length(), element.length());
		}

		try {
			return Long.parseLong(sb.toString());
		}
		catch(NumberFormatException e) {
			return -1;
		}
	}

	private static void logError(Debug logger, String msg) {
		if(logger != null) {
			logger.error(msg);
		}
		else {
			System.out.println("ERROR: " + msg);
		}
	}

	private static void logFine(Debug logger, String msg) {
		if(logger != null) {
			logger.fine(msg);
		}
		else {
			System.out.println("FINE: " + msg);
		}
	}

	private static void logFiner(Debug logger, String msg) {
		if(logger != null) {
			logger.finer(msg);
		}
		else {
			System.out.println("FINER: " + msg);
		}
	}

	private static void logFinest(Debug logger, String msg) {
		if(logger != null) {
			logger.finest(msg);
		}
		else {
			System.out.println("FINEST: " + msg);
		}
	}
}
