/*
 *
 * Copyright 2018-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence;

import com.google.common.collect.ImmutableList;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentNode;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKindCombination;
import de.bsvrz.dav.daf.main.archive.ArchiveDataSpecification;
import de.bsvrz.sys.funclib.kappich.annotations.Nullable;
import java.io.File;
import java.io.IOException;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.stream.Collectors;

/**
 * Klasse, die zu einem Containerverzeichnis die Datenidentifikation und Archivdatenart speichert, damit diese Information zusammen
 * gespeichert werden kann.
 *
 * @author Kappich Systemberatung
 */
public class DataIdentificationDir {

	/**
	 * Verzeichnis
	 */
	private final Path _dataIdentificationDir;

	/**
	 * Datenidentifikation
	 */
	private final IdDataIdentification _dataIdentification;

	/**
	 * Ursprüngliche Archivdatenspezifikation (oder null)
	 */
	private Collection<ArchiveDataSpecification> _ads;

	/**
	 * Datenknoten mit Parametrierung (kann null sein, falls unbekannt). null heißt nicht, dass es keinen gibt, nur dass der Aufrufer diesen ggf. selbst ermitteln muss.
	 */
	@Nullable
	private final DataIdentNode _dataIdentNode;

	/**
	 * Erstellt ein neues DataIdentificationDir-Objekt
	 * @param dataIdentificationDir Verzeichnis
	 * @param dataIdentification Datenidentifikation
	 * @param dataIdentNode
	 */
	public DataIdentificationDir(Path dataIdentificationDir, IdDataIdentification dataIdentification, final DataIdentNode dataIdentNode) {
		_dataIdentificationDir = dataIdentificationDir;
		_dataIdentification = dataIdentification;
		_dataIdentNode = dataIdentNode;
	}

	/**
	 * Erstellt ein neues DataIdentificationDir-Objekt
	 * @param dataIdentificationDir Verzeichnis
	 */
	public DataIdentificationDir(Path dataIdentificationDir) throws PersistenceException {
		this(dataIdentificationDir, PersistenceManager.getDataIdentificationFromPath(dataIdentificationDir.toFile()), null);
	}

	/**
	 * Erstellt ein neues ContainerFileDir-Objekt
	 * @param persistenceManager PersistenceManager
	 * @param dataIdentification Datenidentifikation
	 * @param dataIdentNode      Datenknoten
	 */
	public DataIdentificationDir(PersistenceManager persistenceManager, IdDataIdentification dataIdentification, final DataIdentNode dataIdentNode) {
		this(persistenceManager.getPath(dataIdentification), dataIdentification, dataIdentNode);
	}

	/**
	 * Erstellt ein neues ContainerFileDir-Objekt
	 * @param archiveDataSpecifications Datenspezifikationen
	 * @param persistenceManager PersistenceManager
	 * @param dataIdentification Datenidentifikation
	 */
	public DataIdentificationDir(final PersistenceManager persistenceManager, final IdDataIdentification dataIdentification, final Collection<ArchiveDataSpecification> archiveDataSpecifications) {
		this(persistenceManager, dataIdentification, (DataIdentNode)null);
		_ads = ImmutableList.copyOf(archiveDataSpecifications);
	}

	/**
	 * Erstellt ein neues ContainerFileDir-Objekt
	 * @param persistenceManager PersistenceManager
	 * @param dataIdentNode DataIdentNode
	 */
	public DataIdentificationDir(final PersistenceManager persistenceManager, final DataIdentNode dataIdentNode) {
		this(persistenceManager, dataIdentNode.getDataIdentification(), dataIdentNode);
	}

	/**
	 * Gibt das Verzeichnis zurück
	 * @return das Verzeichnis
	 */
	public Path getDataIdentificationDir() {
		return _dataIdentificationDir;
	}

	/**
	 * Gibt die Datenidentifikation zurück
	 * @return die Datenidentifikation
	 */
	public IdDataIdentification getDataIdentification() {
		return _dataIdentification;
	}
	/**
	 * Gibt die Archivdatenspezifikation zurück (falls sie beim Konstruktor übergeben wurde)
	 * @return Archivdatenspezifikation oder null
	 */
	@Nullable
	public Collection<ArchiveDataSpecification> getArchiveDataSpecifications() {
		return _ads;
	}

	@Override
	public String toString() {
		return _dataIdentification.toString();
	}

	public ArchiveDataKindCombination getArchiveDataKinds() throws PersistenceException {
		if(_ads != null) {
			List<ArchiveDataKind> dataKinds = 
					_ads.stream()
					.flatMap(it -> it.getDataKinds().stream())
					.collect(Collectors.toList());
			
			return new ArchiveDataKindCombination(dataKinds);
		}
		try(DirectoryStream<Path> paths = Files.newDirectoryStream(_dataIdentificationDir)) {
			final List<ArchiveDataKind> dataKindList = new ArrayList<>();
			for(Path path : paths) {
				File dir = path.toFile();
				if(PersistenceManager.isValidDataKindSuffix(dir.getName())) {
					dataKindList.add(PersistenceManager.getArchiveDataKindFromDir(dir));
				}
			}
			return new ArchiveDataKindCombination(dataKindList);
		}
		catch(IOException e) {
			throw new PersistenceException(e);
		}
	}

	/**
	 * Datenknoten mit Parametrierung (kann null sein, falls unbekannt). null heißt nicht, dass es keinen gibt, nur dass der Aufrufer diesen ggf. selbst ermitteln muss.
	 * 
	 * @return DataIdentNode
	 */
	@Nullable
	public DataIdentNode getDataIdentNode() {
		return _dataIdentNode;
	}
}
