/*
 * Copyright 2019 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence;

import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.ArchiveDataSpecification;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.dataIdentificationSettings.DataIdentification;
import de.bsvrz.sys.funclib.kappich.annotations.NotNull;

import java.util.Objects;

/**
 * Klasse, die die IDs der Objekte einer Datenidentifikation aufnehmen kann.
 * @author Kappich Systemberatung
 */
public class IdContainerFileDir implements Comparable<IdContainerFileDir> {

	/** Id des Systemobjekts dieser Datenidentifikation */
	private final long _objectId;

	/** Id der Attributgruppe dieser Datenidentifikation */
	private final long _atgId;

	/** Id des Aspekts dieser Datenidentifikation */
	private final long _aspectId;

	/** Simulationsvariante dieser Datenidentifikation */
	private final int _simVariant;
	
	/** Archivdatenart */
	private final ArchiveDataKind _archiveDataKind;

	/**
	 * Erzeugt ein neues Objekt mit den angegebenen Werten
	 * @param object Objekt
	 * @param dataDescription {@link DataDescription}
	 * @param archiveDataKind Archivdatenart
	 */
	public IdContainerFileDir(final SystemObject object, DataDescription dataDescription, final ArchiveDataKind archiveDataKind) {
		_objectId = object.getId();
		_atgId = dataDescription.getAttributeGroup().getId();
		_aspectId = dataDescription.getAspect().getId();
		_simVariant = dataDescription.getSimulationVariant();
		_archiveDataKind = Objects.requireNonNull(archiveDataKind);
	}

	/**
	 * Erzeugt ein neues Objekt mit den angegebenen Werten
	 *  @param objectId          Id des Systemobjekts dieser Datenidentifikation
	 * @param atgId             Id der Attributgruppe dieser Datenidentifikation
	 * @param aspectId          Id des Aspekts dieser Datenidentifikation
	 * @param simulationVariant Simulationsvariante dieser Datenidentifikation
	 * @param archiveDataKind  Archivdatenart
	 */
	public IdContainerFileDir(final long objectId, final long atgId, final long aspectId, final int simulationVariant, final ArchiveDataKind archiveDataKind) {
		_objectId = objectId;
		_atgId = atgId;
		_aspectId = aspectId;
		_simVariant = simulationVariant;
		_archiveDataKind = Objects.requireNonNull(archiveDataKind);
	}

	/**
	 * Erstellt eine neue Instanz basiert auf einer {@link ArchiveDataSpecification}
	 * @param ads {@link ArchiveDataSpecification}
	 * @param archiveDataKind  Archivdatenart
	 */
	public IdContainerFileDir(final ArchiveDataSpecification ads, final ArchiveDataKind archiveDataKind) {
		this(ads.getObject(), ads.getDataDescription(), archiveDataKind);
	}

	/**
	 * Erstellt eine neue Instanz aus einer {@link DataIdentification}
	 * @param dataIdentification {@link DataIdentification}
	 * @param archiveDataKind  Archivdatenart
	 */
	public IdContainerFileDir(final DataIdentification dataIdentification, final ArchiveDataKind archiveDataKind) {
		this(dataIdentification.getObject(), dataIdentification.getDataDescription(), archiveDataKind);
	}
	
	/**
	 * Erstellt eine neue Instanz aus einer {@link DataIdentification}
	 * @param dataIdentification {@link DataIdentification}
	 * @param archiveDataKind   Archivdatenart
	 */
	public IdContainerFileDir(final IdDataIdentification dataIdentification, final ArchiveDataKind archiveDataKind) {
		this(dataIdentification.getObjectId(), dataIdentification.getAtgId(), dataIdentification.getAspectId(), dataIdentification.getSimVariant(), archiveDataKind);
	}

	@Override
	public boolean equals(final Object o) {
		if(this == o) return true;
		if(o == null || getClass() != o.getClass()) return false;

		final IdContainerFileDir that = (IdContainerFileDir) o;

		if(_objectId != that._objectId) return false;
		if(_atgId != that._atgId) return false;
		if(_aspectId != that._aspectId) return false;
		if(_simVariant != that._simVariant) return false;
		return _archiveDataKind.equals(that._archiveDataKind);
	}

	@Override
	public int hashCode() {
		int result = (int) (_objectId ^ (_objectId >>> 32));
		result = 31 * result + (int) (_atgId ^ (_atgId >>> 32));
		result = 31 * result + (int) (_aspectId ^ (_aspectId >>> 32));
		result = 31 * result + _simVariant;
		result = 31 * result + _archiveDataKind.hashCode();
		return result;
	}

	/**
	 * Gibt die Objekt-ID zurück
	 * @return die Objekt-ID
	 */
	public long getObjectId() {
		return _objectId;
	}

	/**
	 * Gibt die Attributgruppen-ID zurück
	 * @return die Attributgruppen-ID
	 */
	public long getAtgId() {
		return _atgId;
	}

	/**
	 * Gibt die Aspekt-ID zurück
	 * @return die Aspekt-ID
	 */
	public long getAspectId() {
		return _aspectId;
	}

	/**
	 * Gibt die Simulationsvariante zurück
	 * @return die Simulationsvariante
	 */
	public int getSimVariant() {
		return _simVariant;
	}

	/** 
	 * Gibt die Archivdatenart zurück
	 * @return die Archivdatenart
	 */
	public ArchiveDataKind getArchiveDataKind() {
		return _archiveDataKind;
	}

	/**
	 * Gibt die Datenidentifikation zurück
	 * @return Datenidentifikation
	 */
	public IdDataIdentification getDataIdentification() {
		return new IdDataIdentification(_objectId, _atgId, _aspectId, _simVariant);
	}

	public String toString() {
		return "obj: " + _objectId + " atg: " + _atgId + " asp: " + _aspectId + " sv: " + _simVariant + " adk: " + _archiveDataKind;
	}

	@Override
	public int compareTo(@NotNull final IdContainerFileDir o) {
		int tmp;
		tmp = Long.compare(getObjectId(), o.getObjectId());
		if(tmp != 0) return tmp;
		tmp = Long.compare(getAtgId(), o.getAtgId());
		if(tmp != 0) return tmp;
		tmp = Long.compare(getAspectId(), o.getAspectId());
		if(tmp != 0) return tmp;
		tmp = Integer.compare(getSimVariant(), o.getSimVariant());
		if(tmp != 0) return tmp;
		return Integer.compare(getArchiveDataKind().getCode(), o.getArchiveDataKind().getCode());
	}
}
