/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.datatree.synchronization.SyncKey;
import de.bsvrz.ars.ars.mgmt.tasks.Task;
import de.bsvrz.ars.ars.persistence.index.result.IndexResult;
import de.bsvrz.ars.ars.persistence.walk.DataIdentificationDirWalk;
import de.bsvrz.ars.ars.persistence.walk.internal.StatusPrinter;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.sys.funclib.kappich.annotations.NotNull;
import de.bsvrz.sys.funclib.losb.util.Util;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.time.Duration;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;

/**
 * Diese Klasse teilt das Persistenzverzeichnis in disjunkte Teilbaeume auf und stellt alle Verwaltungsinformationen wieder her.
 *
 * @author beck et al. projects GmbH
 * @author Alexander Schmidt
 * @version $Revision$ / $Date$ / ($Author$)
 * @see PersistenceManager#REBUILDINDEX_FILE_FLAG_NAME
 */
public final class RestorePersDirTsk extends Task {

	/** damit man nach dem cleanup noch zu testzwecken auf die ergebnisse zugreifen kann: */
	public static long test_totalDID;
	public static long test_badDID;
	public static long test_goodDID;
	public static long test_viscont;

	private final AtomicLong maxArchiveTime = new AtomicLong();
	private final AtomicLong maxContID = new AtomicLong();

	private final AtomicInteger maxMedienID = new AtomicInteger();

	private final AtomicLong totalDID = new AtomicLong();
	private final AtomicLong badDID = new AtomicLong();
	private final AtomicLong goodDID = new AtomicLong();        // Zähler für DID

	/** Zähler für unerwartete RuntimeExceptions */
	private final AtomicLong _errorCount = new AtomicLong();

	private final AtomicLong containerRead = new AtomicLong();
	private final AtomicLong containerIgnored = new AtomicLong();
	private final AtomicLong indexCreated = new AtomicLong();
	private final AtomicLong indexDeleted = new AtomicLong();

	/** Gibt den benutzten Speicher frei (insbes. den von den einzelnen Tasks zum Wiederherstellen der Indexe benoetigten Speicher) */
	private void cleanUpTask() {
		// damit man nach dem cleanup noch zu testzwecken auf die ergebnisse zugreifen kann:
		test_totalDID = getTotalDID();
		test_badDID = getBadDID();
		test_goodDID = getGoodDID();
		test_viscont = getContainerRead();
	}

	public long getMaxArchiveTime() {
		return maxArchiveTime.get();
	}

	public long getMaxContID() {
		return maxContID.get();
	}

	public int getMaxMedienID() {
		return maxMedienID.get();
	}

	public long getTotalDID() {
		return totalDID.get();
	}

	public long getBadDID() {
		return badDID.get();
	}

	public long getGoodDID() {
		return goodDID.get();
	}

	public long getContainerRead() {
		return containerRead.get();
	}

	public long getContainerIgnored() {
		return containerIgnored.get();
	}
	
	public long getIndexCreated() {
		return indexCreated.get();
	}
	
	public long getIndexDeleted() {
		return indexDeleted.get();
	}

	public long getErrorCount() {
		return _errorCount.get();
	}

	/** @return Wahr, falls kein Verzeichnis nicht wiederhergestellt werden konnte, falsch sonst */
	public boolean success() {
		return getBadDID() == 0 && getErrorCount() == 0;
	}

	public RestorePersDirTsk(ArchiveManager archiveMgr) {
		super(archiveMgr);
		start();
	}

	/** Durchlaueft mit Hilfe eines {@link DataIdentificationDirWalk} jedes Verzeichnis und stellt die Verwaltungsinformation wieder her. */
	@Override
	public void execute() {
		try {
			DataIdentificationDirWalk containerDirWalk = DataIdentificationDirWalk.allDirectories(persMgr);
			containerDirWalk.setStatusPrinter(new StatusPrinter() {
				@Override
				public String getStatusMessage(final String actionName, final Duration runtime, final boolean indexFinished, final long indexedContainerDirs, final long visitedContainerDirs, final long visitedContainerFiles) {
					long goodDirs = getGoodDID();
					long badDirs = getBadDID();
					long sum = goodDirs + badDirs;
					return super.getStatusMessage(actionName, runtime, indexFinished, indexedContainerDirs, visitedContainerDirs, visitedContainerFiles) + "\n"
							+ getProgressMessage(goodDirs, badDirs, sum);
				}

				@Override
				public String getSuccessMessage(final String actionName, final Duration runtime, final long visitedContainerDirs, final long visitedContainerFiles) {
					long goodDirs = getGoodDID();
					long badDirs = getBadDID();
					long sum = goodDirs + badDirs;
					return super.getSuccessMessage(actionName, runtime, visitedContainerDirs, visitedContainerFiles) + "\n"
                           + getProgressMessage(goodDirs, badDirs, sum);
				}
			});
			containerDirWalk.execute(
					"Wiederherstellungslauf",
					archMgr.getNumCheckPersistenceThreads(),
					(dataIdentificationDir, standardContainerDirWalk) -> {
						Path didDir = dataIdentificationDir.getDataIdentificationDir();
						IdDataIdentification dataIdentification = dataIdentificationDir.getDataIdentification();
						if(shouldTerminate()) {
							containerDirWalk.terminate();
						}
						try(SyncKey<IdDataIdentification> lock = archMgr.lockIndex(dataIdentification)) {
							for(ArchiveDataKind archiveDataKind : dataIdentificationDir.getArchiveDataKinds()) {
								try {
									totalDID.getAndIncrement();

									// Indexe ggf. wiederherstellen:
									Path changedFlagFile = persMgr.getRebuildIdxFile(dataIdentificationDir, archiveDataKind);

									// Strategie: Wenn die _rebuildIndex.flag-Datei vorhanden ist, werden die Indexe neu aufgebaut.
									boolean needsRebuild = Files.exists(changedFlagFile);

									if(needsRebuild) {
										long t2 = Util.startTimer();
										PersistenceManager.RebuildResult result = new PersistenceManager.RebuildResult();
										persMgr.rebuildIndexes(lock, archiveDataKind, result);
										containerRead.addAndGet(result.getContainerRead());
										containerIgnored.addAndGet(result.getContainerIgnored());
										indexCreated.addAndGet(result.getIndexCreated());
										indexDeleted.addAndGet(result.getIndexDeleted());
										try {
											Files.deleteIfExists(changedFlagFile);
										}
										catch(IOException e) {
											badDID.getAndIncrement();
											throw new IOException(
													"Wiederherstellungslauf: '" + PersistenceManager.REBUILDINDEX_FILE_FLAG_NAME + "' im Verzeichnis '" + didDir
															+ "' konnte nicht geloescht werden.", e
											);
										}
										goodDID.getAndIncrement();
										_debug.fine("Wiederherstellungslauf: Indexe in " + didDir + " erfolgreich neu aufgebaut (" + Util.stopTimer(t2) + ").");

									}

									// maximale Archivzeit. Medien- und Container-ID ermitteln
									IndexResult<ContainerManagementInformation> headers = persMgr.getContainerHeaders(lock, archiveDataKind);
									if(!headers.isEmpty()) {
										long contId = headers.getMax(ContainerManagementInformation.CHP_CONT_ID);
										long aTime = persMgr.maxATime(new IdContainerFileDir(dataIdentification, archiveDataKind));
										int medienID = persMgr.maxMedienID(headers);

										maxContID.updateAndGet(old -> Math.max(old, contId));
										maxArchiveTime.updateAndGet(old -> Math.max(old, aTime));
										maxMedienID.updateAndGet(old -> Math.max(old, medienID));
									}
								}
								catch(Exception e) {
									badDID.getAndIncrement();
									_debug.error("Wiederherstellungslauf: Index der Datenidentifikation '" + didDir + "' konnte nicht neu angelegt werden", e);
								}
							}
						}
					}
			);
		}
		catch(PersistenceException e) {
			_errorCount.getAndIncrement();
			_debug.error("Wiederherstellungslauf: Unbehandelter Fehler in RestorePersDirTsk", e);
		}
		finally {
			cleanUpTask();
			super.execute();
		}
	}

	@NotNull
	private String getProgressMessage(final long goodDirs, final long badDirs, final long sum) {
		final String result;
		if(sum != 0) {
			StringBuilder sb = new StringBuilder().append("bei ")
					                                     .append(sum)
					                                     .append(" Index-Neuaufbau notwendig, davon ")
					                                     .append(goodDirs)
					                                     .append(" erfolgreich und ")
					                                     .append(badDirs)
					                                     .append(" nicht erfolgreich.")
					                                     .append("\n");
			if(getContainerRead() != 0) {
                sb.append(String.format("%8d Container erfolgreich gelesen.\n", getContainerRead()));
			}
			if(getContainerIgnored() != 0) {
                sb.append(String.format("%8d defekte Container ignoriert.\n", getContainerIgnored()));
			}	
			if(getIndexCreated() != 0) {
                sb.append(String.format("%8d Indexdateien erstellt.\n", getIndexCreated()));
			}	
			if(getIndexDeleted() != 0) {
                sb.append(String.format("%8d Indexdateien gelöscht.\n", getIndexDeleted()));
			}
			result = sb.toString();
		}
		else {
			result = "kein Index-Neuaufbau notwendig.";
		}
		return result;
	}

}
