/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence.index;

import com.google.common.collect.ImmutableMap;
import de.bsvrz.ars.ars.persistence.index.backend.management.AbstractIndex;
import de.bsvrz.ars.ars.persistence.index.backend.management.ColumnType;
import de.bsvrz.ars.ars.persistence.index.backend.management.IndexContentDescriptor;
import de.bsvrz.ars.ars.persistence.index.backend.management.LongRange;
import de.bsvrz.ars.ars.persistence.index.result.IndexResult;
import java.nio.file.Path;

/**
 * Diese Klasse stellt einen DatenzeitIndex dar. Sie verhält sich ähnlich dem {@link ArchiveTimeIndexImpl}. Der Unterschied ist, dass die Folge der Eintraege
 * (min, max, min, max...) nicht monoton ist, weil es Rückspruenge geben kann. Stattdessen sind die Eintraege nur nach dem min-Wert sortiert.
 *
 * @author beck et al. projects GmbH
 * @author Phil Schrettenbrunner
 * @version $Revision$ / $Date$ / ($Author$)
 */
public final class DataTimeIndexImpl extends AbstractIndex<IndexValues> implements DataTimeIndex {

	/** Dateiname, unter dem alle Indices dieser Art gespeichert werden */
	public static final String IDX_FILENAME = "_datatime.idx";

	private static final IndexContentDescriptor<IndexValues> INDEX_CONTENT_DESCRIPTOR;

	static {
		INDEX_CONTENT_DESCRIPTOR = new IndexContentDescriptor<>(IndexValues.class);
		INDEX_CONTENT_DESCRIPTOR.addRange(6, IndexValues.DataTimeMin, ColumnType.Increasing, IndexValues.DataTimeMax, ColumnType.Unordered);
		INDEX_CONTENT_DESCRIPTOR.addRange(8, IndexValues.DataIndexMin, ColumnType.Unordered, IndexValues.DataIndexMax, ColumnType.Unordered);
		INDEX_CONTENT_DESCRIPTOR.addColumn(5, IndexValues.ContainerId, ColumnType.Unique);
	}


	/**
	 * Erzeugt einen neuen Datenzeitindex.
	 * <p>
	 * Bei startmode==APPEND_REUSE_LAST wird der Ringpuffer mit den aktuellsten Daten aus einer vorhandenen Indexdatei gefüllt. Der Eintrag mit der größten
	 * ContainerId wird als currentEntry reaktiviert. Damit kann ein Container fortgesetzt und der reaktivierte Eintrag weiter aktualisiert werden.
	 * <p>
	 * Bei startmode==ERASE_OLD wird ein leerer Index erzeugt und die bisherige Indexdatei (sofern vorhanden) geloescht.
	 *
	 * @param bufferSize Anzahl der Einträge, die der die der Index im RAM halten soll
	 * @param file Datei
	 *
	 * @throws CorruptIndexException Bei Problemen mit dem Dateizugriff
	 */
	public DataTimeIndexImpl(int bufferSize, Path file) throws CorruptIndexException {
		super(INDEX_CONTENT_DESCRIPTOR, bufferSize, file);
	}
	
	////////////////////////////// Methoden auf den Current*-Werten ////////////////////////////

	/**
	 * Fügt dem Index einen neuen Eintrag hinzu. Dabei wird der derzeitig aktuelle Eintrag abgeschlossen. Führt dies dazu, dass keine weiteren freien Plätze im
	 * Cache vorhanden sind, werden die Änderungen auf Platte geschrieben. Wenn der Cache noch vollständig leer ist, werden die aktuellsten Daten aus der
	 * Index-Datei geladen.
	 *
	 * @param dTMin    DatenZeit Min-Wert
	 * @param dTMax    DatenZeit Max-Wert
	 * @param dIMin    DatenIndex Min-Wert
	 * @param dIMax    DatenIndex Max-Wert
	 * @param contId   Container ID
	 *
	 * @throws IndexException Bei Problemem mit der Index-Datei
	 */
	public void add(long dTMin, long dTMax, long dIMin, long dIMax, long contId) throws IndexException {
		checkAdd(dTMin, dTMax, dIMin, dIMax, contId);

		setInsertValue(IndexValues.DataIndexMin, dIMin);
		setInsertValue(IndexValues.DataIndexMax, dIMax);
		setInsertValue(IndexValues.DataTimeMin, dTMin);
		setInsertValue(IndexValues.DataTimeMax, dTMax);
		setInsertValue(IndexValues.ContainerId, contId);
		insert();
	}


	private void checkAdd(final long dTMin, final long dTMax, final long dIMin, final long dIMax, final long contId) throws IndexException {
		if(dIMin < 0 || dIMax < 0 || dTMin < 0 || dTMax < 0 || contId <= 0) {
			throw new CorruptIndexException(
					"DatenzeitIndex: neuer Eintrag ungültig: (" + "dIMin=" + dIMin + ", dIMax=" + dIMax + ", dTMin=" + dTMin + ", dTMax=" + dTMax + ", contID="
							+ contId + ")", getFile()
			);
		}
//		if(getCurrent(IndexValues.ContainerId) >= contId) {
//			throw new CorruptIndexException("DatenzeitIndex: ContainerId (" + contId + ") des neuen Eintrags ungültig: currentCId=" + getCurrent(IndexValues.ContainerId), getFile());
//		}
	}

	/**
	 * Liefert eine Matrix mit DTMin/Max, DIMin/Max und ContainerIds, deren DTMin/Max-Werte zwischen dTMin und dTMax liegen. Wird diese Methode aufgerufen wenn der
	 * Cache noch leer ist, werden die aktuellsten Daten aus der Datei nachgeladen.
	 *
	 * @param dTMin    Min-Wert der Zeitspanne
	 * @param dTMax    Max-Wert der Zeitspanne
	 *
	 * @return IndexResult-Objekt, das die Index innerhalb der gesuchten Zeitspanne enthält (DT und DI Werte, keine AT Werte)
	 *
	 * @throws IndexException Bei Problemem mit der IndexDatei
	 */
	@Override
	public IndexResult<IndexValues> getContainerIDByDataTime(long dTMin, long dTMax) throws IndexException {
		// Der DatenZeitMaximum des Containers muss größer als das Anfrageminimum sein und der DatenZeitMinimum des Containers muss kleiner als das AnfrageMaximum sein.
		// Spezialfälle, wie dass ggf. der Container vor dem Anfrageminimum zurückgeliefert muss, werden von der Implementierung beachtet.
		return query(ImmutableMap.of(IndexValues.DataTimeMax, new LongRange(dTMin, null), IndexValues.DataTimeMin, new LongRange(null, dTMax)));
	}

	/**
	 * Gibt die anzahl Bytes zurück, die ein Indexeintrag belegt.
	 * @return Anzahl bytes
	 */
	public static int entrySize() {
		return INDEX_CONTENT_DESCRIPTOR.getEntryLengthBytes();
	}
}
