/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009-2019 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence.index;

import de.bsvrz.sys.funclib.losb.util.ByteIO;

/**
 * Hilfsklasse für einen Datenzeit Index.
 * <p>
 * Ein Eintrag im Datenzeit Index ist wie folgt aufgebaut: <table><caption>Datenindex-Aufbau</caption><tr> <td>DT-Min</td><td>DT-Max</td><td>DI-Min</td><td>DI-Max</td><td>CID</td> </tr> <tr>
 * <td>6 Byte</td><td>6 Byte</td><td>8 Byte</td><td>8 Byte</td><td>5 Byte</td> </tr> </table>
 * <p>
 * Die Gesamtlänge ist 33 Bytes. Ein Eintrag wird mit {@link #setEntry(byte[], int, long[])} geschrieben und über die getter ausgelesen.
 *
 * @author beck et al. projects GmbH
 * @author Thomas Schaefer
 * @author Philippe Schrettenbrunner
 * @version $Revision$ / $Date$ / ($Author$)
 */
@Deprecated
public class IndexEntryTime {

	private static final int DT_MIN_POS = 0;                              // 6 Byte Länge

	private static final int DT_MAX_POS = DT_MIN_POS + ByteIO.LONG6B_LEN; // 6 Byte Länge

	private static final int DI_MIN_POS = DT_MAX_POS + ByteIO.LONG6B_LEN; // 8 Byte Länge

	private static final int DI_MAX_POS = DI_MIN_POS + ByteIO.LONG8B_LEN;   // 8 Byte Länge

	private static final int CONTID_POS = DI_MAX_POS + ByteIO.LONG8B_LEN;   // 5 Byte Länge

	/** Länge eines Eintrages im Index */
	public static final int ENTRY_LEN = CONTID_POS + ByteIO.LONG5B_LEN;


	/** Rein statische Klasse, keine Instanzen -> privater Konstruktor */
	private IndexEntryTime() {
	}


	/**
	 * Schreibt einen Eintrag an die Stelle pos im ByteBuffer buffer
	 *
	 * @param buffer Puffer, in den der Eintrag geschrieben wird
	 * @param pos    Index der Stelle, an der der Eintrag geschrieben wird
	 * @param entry  Long Array mit Min-Wert der Datenzeit, Max-Wert der Datenzeit, Min-Wert des DatenIndex, Max-Wert des DatenIndexund der Container id
	 */
	public static void setEntry(byte[] buffer, int pos, long[] entry) {
		assert buffer != null : "Kein Puffer";
		assert pos * ENTRY_LEN < buffer.length : "Ungültige Position";
		assert entry.length == 5 : "Länge Entry Array != 5";

		int startPos = ENTRY_LEN * pos;
		ByteIO.writeUnsignedLong6Byte(buffer, startPos + DT_MIN_POS, entry[0]);
		ByteIO.writeUnsignedLong6Byte(buffer, startPos + DT_MAX_POS, entry[1]);
		ByteIO.writeSignedLong8Byte(buffer, startPos + DI_MIN_POS, entry[2]);
		ByteIO.writeSignedLong8Byte(buffer, startPos + DI_MAX_POS, entry[3]);
		ByteIO.writeUnsignedLong5Byte(buffer, startPos + CONTID_POS, entry[4]);
	}


	/**
	 * Liefert den gesammten Eintrag an der Stelle pos im Puffer buffer.
	 *
	 * @param buffer Byte Buffer
	 * @param pos    Index des Eintrags, der gelesen werden soll
	 *
	 * @return Array mit DTMin, DTMax, DIMin, DIMax, CID
	 */
	public static long[] getEntry(byte[] buffer, int pos) {
		return new long[]{
				getDTMin(buffer, pos), getDTMax(buffer, pos), getDIMin(buffer, pos), getDIMax(buffer, pos), getCID(buffer, pos)
		};
	}


	/**
	 * Liefert den Min-Wert der Datenzeit.
	 *
	 * @param buffer Byte Buffer
	 * @param pos    Index des Eintrags, der gelesen werden soll
	 *
	 * @return Min-Wert der Datenzeit
	 */
	public static long getDTMin(byte[] buffer, int pos) {
		return ByteIO.readUnsignedLong6Byte(buffer, pos * ENTRY_LEN + DT_MIN_POS);
	}


	/**
	 * Liefert den Max-Wert der Datenzeit.
	 *
	 * @param buffer Byte Buffer
	 * @param pos    Index des Eintrags, der gelesen werden soll
	 *
	 * @return Max-Wert der Datenzeit
	 */
	public static long getDTMax(byte[] buffer, int pos) {
		return ByteIO.readUnsignedLong6Byte(buffer, pos * ENTRY_LEN + DT_MAX_POS);
	}


	/**
	 * Liefert den Min-Wert des Datenindex.
	 *
	 * @param buffer Byte Buffer
	 * @param pos    Index des Eintrags, der gelesen werden soll
	 *
	 * @return Min-Wert des Datenindex
	 */
	public static long getDIMin(byte[] buffer, int pos) {
		return ByteIO.readSignedLong8Byte(buffer, pos * ENTRY_LEN + DI_MIN_POS);
	}


	/**
	 * Liefert den Max-Wert des Datenindex.
	 *
	 * @param buffer Byte Buffer
	 * @param pos    Index des Eintrags, der gelesen werden soll
	 *
	 * @return Max-Wert des Datenindex
	 */
	public static long getDIMax(byte[] buffer, int pos) {
		return ByteIO.readSignedLong8Byte(buffer, pos * ENTRY_LEN + DI_MAX_POS);
	}


	/**
	 * Liefert die Container-ID.
	 *
	 * @param buffer Byte Buffer
	 * @param pos    Index des Eintrags, der gelesen werden soll
	 *
	 * @return die Container-ID
	 */
	public static long getCID(byte[] buffer, int pos) {
		return ByteIO.readUnsignedLong5Byte(buffer, pos * ENTRY_LEN + CONTID_POS);
	}
}
