/*
 *
 * Copyright 2009-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.repair;

import de.bsvrz.ars.ars.mgmt.datatree.synchronization.SynchronizationFailedException;
import de.bsvrz.ars.ars.persistence.BasicContainerFileHandle;
import de.bsvrz.ars.ars.persistence.ContainerFile;
import de.bsvrz.ars.ars.persistence.ContainerHdr;
import de.bsvrz.ars.ars.persistence.ContainerManagementInformation;
import de.bsvrz.ars.ars.persistence.DeletedContainerFile;
import de.bsvrz.ars.ars.persistence.IdContainerFileDir;
import de.bsvrz.ars.ars.persistence.IdDataIdentification;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.ars.ars.persistence.PersistenceManager;
import de.bsvrz.ars.ars.persistence.index.ArchiveTimeIndexImpl;
import de.bsvrz.ars.ars.persistence.index.ContainerManagementIndex;
import de.bsvrz.ars.ars.persistence.index.DataIndexAndArchiveTimeIndex;
import de.bsvrz.ars.ars.persistence.index.DataIndexIndexImpl;
import de.bsvrz.ars.ars.persistence.index.DataTimeIndexImpl;
import de.bsvrz.ars.ars.persistence.index.IndexException;
import de.bsvrz.ars.ars.persistence.index.result.IndexResult;
import de.bsvrz.ars.ars.persistence.iter.DataIterator;
import de.bsvrz.ars.ars.persistence.walk.DataIdentificationDirWalk;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.debug.Debug;
import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.regex.Pattern;

/**
 * Hilfsprogramm, mit dem es möglich ist, Container-Dateien des Archivsystems, in denen bestimmte Defekte vorliegen zu identifizieren und zu korrigieren. Das
 * Programm durchsucht in einem vorgegebenen Verzeichnis und auf Wunsch auch rekursiv in allen Unterverzeichnisse nach Container-Dateien und führt Prüfungen auf
 * bestimmte Defekte durch, die durch Fehler in älteren Versionen des Archivsystems ausgelöst wurden. Defekte Containerdateien werden durch eine entsprechende
 * Ausgabe identifiziert und auf Wunsch wenn möglich auch korrigiert.
 * <p>
 * Das Programm unterstützt folgende Aufrufparameter: -verzeichnis=... -rekursiv=ja|nein -reparieren=nein|ja -index=nein|ja -container=nein|ja
 * <p>
 * Mit {@code -verzeichnis=...} kann ein beliebiges Verzeichnis angegeben werden, in dem Containerdateien gesucht werden. Defaultwert ist das aktuelle
 * Arbeitsverzeichnis in dem das Programm gestartet wurde.
 * <p>
 * Mit {@code -rekursiv=ja|nein} kann spezifiziert werden, ob auch Unterverzeichnisse des Ausgangsverzeichnisses rekursiv durchsucht werden sollen oder
 * nicht. Ohne Angabe dieses Arguments werden auch Unterverzeichnisse rekursiv durchsucht.
 * <p>
 * Mit {@code -reparieren=nein|ja} kann spezifiziert werden, ob fehlerhafte Dateien nur identifiziert werden sollen oder ob sie auch repariert werden
 * sollen.
 * <p>
 * Mit {@code -index=ja|nein} kann spezifiziert werden, ob Indexdateien überprüft werden
 * sollen.
 * <p>
 * Mit {@code -container=ja|nein} kann spezifiziert werden, ob Containerdateien überprüft werden
 * sollen.
 *
 * @author Kappich Systemberatung
*/
@SuppressWarnings("UseOfSystemOutOrSystemErr")
public class ContainerRescue {

	private final File _startDirectory;

	private final boolean _recursiveScan;

	private final boolean _fixDefects;

	private final boolean _checkContainerFiles;

	private final boolean _checkIndexFiles;

	private final boolean _deleteBackupIndex;

	private final FilenameFilter _containerFilenameFilter;

	private final Pattern _containerFilenamePattern;

	private final FilenameFilter _indexFilenameFilter;

	private final Pattern _indexFilenamePattern;

	private final List<Defect> _cantFix;

	private List<Defect> _oneByteShortDefects;

	private long _validContainerFileCount;

	private long _checkedContainerFileCount;

	private long _defectContainerFileCount;

	private long _validIndexFileCount;

	private long _checkedIndexFileCount;

	private long _defectIndexFileCount;

	private List<Defect> _badRangeEndDefects;

	private List<Defect> _badContainerFileEnd;

	private List<Defect> _shortContainerFiles;

	private List<Defect> _badIndexFiles;

	private long _correctedDefectCount;

	private int _deletedFiles;

	public static void main(String[] args) {
		try {
			final ArgumentList arguments = new ArgumentList(args);
			Debug.init("ContainerRescue", arguments);
			final ContainerRescue containerRescue = new ContainerRescue(arguments);
			containerRescue.start();
		}
		catch(Exception e) {
			e.printStackTrace();
			System.err.println("Ein unerwarteter Fehler ist während der Programmausführung aufgetreten: " + e);
		}
	}

	public ContainerRescue(final ArgumentList arguments) {
		_startDirectory = arguments.fetchArgument("-verzeichnis=.").asDirectory();
		_recursiveScan = arguments.fetchArgument("-rekursiv=ja").booleanValue();
		_checkContainerFiles = arguments.fetchArgument("-container=ja").booleanValue();
		_checkIndexFiles = arguments.fetchArgument("-index=ja").booleanValue();
		_fixDefects = arguments.fetchArgument("-reparieren=nein").booleanValue();
		_deleteBackupIndex = arguments.fetchArgument("-bereinigen=nein").booleanValue();
		arguments.ensureAllArgumentsUsed();
		_cantFix = Collections.synchronizedList(new ArrayList<>());
		_oneByteShortDefects = Collections.synchronizedList(new ArrayList<>());
		_badRangeEndDefects = Collections.synchronizedList(new ArrayList<>());
		_badContainerFileEnd = Collections.synchronizedList(new ArrayList<>());
		_badIndexFiles = Collections.synchronizedList(new ArrayList<>());
		_shortContainerFiles = Collections.synchronizedList(new ArrayList<>());
		_containerFilenamePattern = Pattern.compile("dc[0-9]*\\.dat");
		_containerFilenameFilter = (dir, name) -> _containerFilenamePattern.matcher(name).matches();
		_indexFilenamePattern = Pattern.compile("_.+\\.idx");
		_indexFilenameFilter = (dir, name) -> _indexFilenamePattern.matcher(name).matches();
	}

	private void start() throws PersistenceException {
		System.out.println(
				"Aufrufparameter: -verzeichnis=" + _startDirectory + " -rekursiv=" + (_recursiveScan ? "ja" : "nein") + " -container="
				+ (_checkContainerFiles ? "ja" : "nein") + " -index=" + (_checkIndexFiles ? "ja" : "nein") + " -reparieren=" + (_fixDefects ? "ja" : "nein")
		);
		if(!_checkContainerFiles && !_checkIndexFiles){
			System.out.println("########## Bei -container=nein und -index=nein wird nichts geprüft.");
			return;
		}

		System.out.println(
				"########## Das Verzeichnis " + _startDirectory.getAbsolutePath()
				+ (_recursiveScan ? " und rekursiv alle enthaltenen Unterverzeichnisse werden" : " wird") + " nach defekten "
				+ (_checkContainerFiles ? "Container" : "") + ((_checkContainerFiles && _checkIndexFiles) ? "- und " : "") + (_checkIndexFiles ? "Index" : "")
				+ "-Dateien untersucht. " + "Defekte Dateien werden ausgegeben " + (_fixDefects ? "und" : "aber nicht") + " repariert."
		);
		long t0 = System.currentTimeMillis();
		if(!_recursiveScan) {
			scanDirectory(_startDirectory);
		}
		else {
			DataIdentificationDirWalk walk = DataIdentificationDirWalk.create(_startDirectory.toPath());
			walk.execute("ContainerRescue", 11, (containerFileDir, standardContainerDirWalk) -> {
				             Path dataIdentificationDir = containerFileDir.getDataIdentificationDir();
				             for(ArchiveDataKind dataKind : containerFileDir.getArchiveDataKinds()) {
					             scanDirectory(dataIdentificationDir.resolve(PersistenceManager.getDataKindSuffix(dataKind)).toFile());
				             }
			}
			);
		}
		long t1 = System.currentTimeMillis();
		long scanTime = t1 - t0;
		System.out.println("########## Zusammenfassung der Prüfung:");
		System.out.println("Dauer der Prüfung:                " + String.format("%.3f Sekunden", scanTime / 1000.0));
		System.out.println("Anzahl geprüfter Container:       " + _checkedContainerFileCount);
		System.out.println("Anzahl Container ohne Defekte:    " + _validContainerFileCount);
		System.out.println("Anzahl Container mit Defekten:    " + _defectContainerFileCount);
		System.out.println("Anzahl geprüfter Indexdateien:    " + _checkedIndexFileCount);
		System.out.println("Anzahl Indexdateien ohne Defekte: " + _validIndexFileCount);
		System.out.println("Anzahl Indexdateien mit Defekten: " + _defectIndexFileCount);
		final long fixableDefectsCount =
            _oneByteShortDefects.size() + _badRangeEndDefects.size() + _badContainerFileEnd.size() + _badIndexFiles.size();
		System.out.println("Anzahl korrigierbare Defekte :    " + fixableDefectsCount);
		if(_deleteBackupIndex) {
			System.out.println("Anzahl gelöschte Backup-Index-Dateien :    " + _deletedFiles);
		}
		if(fixableDefectsCount > 0) {
			if(_fixDefects) {
				if(_oneByteShortDefects.size() > 0) {
					System.out.println("########## Korrektur der gelöschten Container, die ein Byte zu kurz sind. Anzahl: " + _oneByteShortDefects.size());
					final List<Defect> oneByteShortDefects = _oneByteShortDefects;
					_oneByteShortDefects = new ArrayList<>();
					for(Defect oneByteShortDefect : oneByteShortDefects) {
						fixOneByteShortDefect(oneByteShortDefect);
					}
					System.out.println("  ########## Erneute Prüfung der korrigierten Container:");
					for(Defect oneByteShortDefect : oneByteShortDefects) {
						checkContainerFile(oneByteShortDefect.getFile());
					}
					final long correctedDefectCount = oneByteShortDefects.size() - _oneByteShortDefects.size();
					System.out.println("  ########## Anzahl erfolgreich korrigierter Defekte: " + correctedDefectCount);
					_correctedDefectCount += correctedDefectCount;
				}
				if(_badRangeEndDefects.size() > 0) {
					System.out.println(
							"########## Korrektur der gelöschten Container, die eine falsche Endangabe eines Datensatzindexbereichs enthalten. Anzahl: "
							+ _badRangeEndDefects.size()
					);
					final List<Defect> badRangeEndDefects = _badRangeEndDefects;
					_badRangeEndDefects = new ArrayList<>();
					for(Defect badRangeEndDefect : badRangeEndDefects) {
						fixBadRangeEndDefect(badRangeEndDefect);
					}
					System.out.println("  ########## Erneute Prüfung der korrigierten Container:");
					for(Defect badRangeEndDefect : badRangeEndDefects) {
						checkContainerFile(badRangeEndDefect.getFile());
					}
					final long correctedDefectCount = badRangeEndDefects.size() - _badRangeEndDefects.size();
					System.out.println("  ########## Anzahl erfolgreich korrigierter Defekte: " + correctedDefectCount);
					_correctedDefectCount += correctedDefectCount;
				}
				if(_badContainerFileEnd.size() > 0) {
					System.out.println(
							"########## Korrektur der Container-Dateien mit ungültigen Bytes am Dateiende. Anzahl: "
							+ _badContainerFileEnd.size()
					);
					final List<Defect> badContainerFileEnds = _badContainerFileEnd;
					_badContainerFileEnd = new ArrayList<>();
					for(Defect badContainerFileEnd : badContainerFileEnds) {
						fixBadContainerFileEnd(badContainerFileEnd);
					}
					System.out.println("  ########## Erneute Prüfung der korrigierten Container:");
					for(Defect badContainerFileEnd : badContainerFileEnds) {
						checkContainerFile(badContainerFileEnd.getFile());
					}
					final long correctedDefectCount = badContainerFileEnds.size() - _badContainerFileEnd.size();
					System.out.println("  ########## Anzahl erfolgreich korrigierter Defekte: " + correctedDefectCount);
					_correctedDefectCount += correctedDefectCount;
				}
				if(_shortContainerFiles.size() > 0) {
					System.out.println(
							"########## Korrektur der leeren Container-Dateien. Anzahl: "
							+ _shortContainerFiles.size()
					);
					final List<Defect> shortContainerFiles = _shortContainerFiles;
					_shortContainerFiles = new ArrayList<>();
					for(Defect shortContainerFile : shortContainerFiles) {
						fixShortContainerFile(shortContainerFile);
					}
					_correctedDefectCount += shortContainerFiles.size();
				}
				if(_badIndexFiles.size() > 0) {
					System.out.println(
							"########## Korrektur von Indexdateien, die fehlerhafte oder unvollständige Einträge enthalten. Anzahl: "
							+ _badIndexFiles.size()
					);
					final List<Defect> badIndexFileDefects = _badIndexFiles;
					_badIndexFiles = new ArrayList<>();
					for(Defect badIndexFileDefect : badIndexFileDefects) {
						fixBadIndexFileDefect(badIndexFileDefect);
					}
					System.out.println("  ########## Erneute Prüfung der korrigierten Indexdateien:");
					for(Defect badIndexFileDefect : badIndexFileDefects) {
						// Container-IDs neu holen, es könnten leere Container gelöscht worden sein
						HashSet<Long> containerIDs = getContainerIdsInDirectory(badIndexFileDefect.getFile().getParentFile());
						checkIndexFile(badIndexFileDefect.getFile(), containerIDs);
					}
					final long correctedDefectCount = badIndexFileDefects.size() - _badIndexFiles.size();
					System.out.println("  ########## Anzahl erfolgreich korrigierter Defekte: " + correctedDefectCount);
					_correctedDefectCount += correctedDefectCount;
				}
				System.out.println("########## Gesamtzahl erfolgreich korrigierter Defekte: " + _correctedDefectCount);
				final long t2 = System.currentTimeMillis();
				long fixTime = t2 - t1;
				System.out.println("Dauer der Korrektur:             " + String.format("%.3f Sekunden", fixTime / 1000.0));
			}
			else {
				System.out.println("########## Zur Korrektur der korrigierbaren Defekte bitte das Aufrufargument -reparieren=ja verwenden.");
			}
		}
	}

	private static void fixShortContainerFile(final Defect defect) {
		File containerFile = defect.getFile();
		containerFile.delete();
		try {
			File indexFile = new File(containerFile.getParentFile(), ContainerManagementIndex.IDX_FILENAME);
			if(indexFile.exists()) {
				indexFile.delete();
			}
		}
		catch(Exception e) {
			System.out.println(containerFile.toString() + ": Unerwarteter Fehler bei der Reparatur: " + e);
		}
	}

	private void scanDirectory(final File directory) {
		try {
			File[] containerFileArray = directory.listFiles(_containerFilenameFilter);
			if(containerFileArray == null) containerFileArray = new File[0];
			File[] containerFiles = containerFileArray;
			if(_checkContainerFiles) {
				for(File containerFile : containerFiles) {
					checkContainerFile(containerFile);
				}
			}
			if(_checkIndexFiles) {
				HashSet<Long> containerIDs = getContainerIdsInDirectory(directory);
				File[] indexFileArray = directory.listFiles(_indexFilenameFilter);
				if(indexFileArray != null) {
					for(File indexFile : indexFileArray) {
						checkIndexFile(indexFile, containerIDs);
					}
				}
			}
		}
		catch(Exception e) {
			System.out.println(directory.toString() + ": Unerwarteter Fehler bei der Reparatur: " + e);
		}
	}

	private HashSet<Long> getContainerIdsInDirectory(final File directory) {
		File[] containerFileArray = directory.listFiles(_containerFilenameFilter);
		if(containerFileArray == null) containerFileArray = new File[0];
		List<File> containerFiles = Arrays.asList(containerFileArray);
		HashSet<Long> result = fileListToIdList(containerFiles);
		try {

			IdDataIdentification identificationFromPath = PersistenceManager.getDataIdentificationFromPath(directory);
			ArchiveDataKind archiveDataKindFromDir = PersistenceManager.getArchiveDataKindFromDir(directory);
			DeletedContainerFile deletedContainerFile = new DeletedContainerFile(directory.toPath(), new IdContainerFileDir(identificationFromPath, archiveDataKindFromDir));
			if(deletedContainerFile.exists()) {
				deletedContainerFile.read();
			}
			result.addAll(deletedContainerFile.containers());
		}
		catch(PersistenceException ignored) {
		}
		return result;
	}

	private static HashSet<Long> fileListToIdList(final List<File> containerFiles) {
		final HashSet<Long> hashSet = new HashSet<>();
		for(File containerFile : containerFiles) {
			hashSet.add(ContainerFile.getContID(containerFile));
		}
		return hashSet;
	}

	private void checkIndexFile(final File indexFile, final HashSet<Long> containerIDs) {
		_checkedIndexFileCount++;
		final Defect defect = inspectIndexFile(indexFile, containerIDs);
		if(defect == null) {
			_validIndexFileCount++;
		}
		else {
			_defectIndexFileCount++;
			defect.setContainerIDsInDirectory(new HashSet<>(containerIDs));
			defect.getCategory().add(defect);
		}
	}

	private Defect inspectIndexFile(final File indexFile, final HashSet<Long> containerIDs) {
		try {
			if(indexFile.canRead()) {
				try(RandomAccessFile raf = new RandomAccessFile(indexFile, "r")) {
					if(indexFile.getName().startsWith("_didx")) {
						return checkDIDIndexFile(raf, indexFile, containerIDs);
					}
					else if(indexFile.getName().startsWith("_datatime")) {
						return checkDataTimeIndexFile(raf, indexFile, containerIDs);
					}
					else if(indexFile.getName().startsWith("_dataindex")) {
						return checkDataIndexFile(raf, indexFile, containerIDs);
					}
					else if(indexFile.getName().startsWith("_backupindex")) {
						return checkBackupIndexFile(raf, indexFile, containerIDs);
					}
					else if(indexFile.getName().equals(ContainerManagementIndex.IDX_FILENAME)) {
						return checkManagementIndexFile(raf, indexFile, containerIDs);
					}
				}
			}
			else {
				return new Defect(indexFile, _cantFix, "Datei nicht lesbar");
			}
		}
		catch(EOFException e) {
			return new Defect(indexFile, _badIndexFiles, "Datei nicht vollständig");
		}
		catch(IOException e) {
			return new Defect(indexFile, _cantFix, "Fehler beim Lesen: " + e);
		}
		return new Defect(indexFile, _cantFix, "Unbekannter Dateityp");
	}

	private Defect checkManagementIndexFile(final RandomAccessFile raf, final File indexFile, final HashSet<Long> containerIDs) {
		try {
			final IndexResult<ContainerManagementInformation> result;
			final HashSet<Long> containerFileIDs;
			try(ContainerManagementIndex index = new ContainerManagementIndex(1024, indexFile.toPath())) {
				result = index.query();
			}
			containerFileIDs = new HashSet<>(containerIDs);

			for(int i = 0; i < result.size(); i++) {
				long contId = result.get(i, ContainerManagementInformation.CHP_CONT_ID);
				containerFileIDs.remove(contId);
			}

			if(containerFileIDs.size() > 0) return new Defect(indexFile, _badIndexFiles, "Es kommen nicht alle ContainerIDs in der Indexdatei vor.");
		}
		catch(IndexException e) {
			return new Defect(indexFile, _badIndexFiles, "ContainerHeaderIndex konnte nicht gelesen werden: " + e.getMessage());
		}
		return null;
	}

	private Defect checkBackupIndexFile(final RandomAccessFile raf, final File indexFile, final HashSet<Long> containerIDs) {

		if(_deleteBackupIndex) {
			if(indexFile.delete()) {
				_deletedFiles++;
			}
		}

		return null;
	}

	/**
	 * Prüft Indexdateien der Klasse {@link ArchiveTimeIndexImpl}. In dieser Klasse ist gefordert, dass die Archivzeit monoton und
	 * die ContainerID streng monoton über alle Eintraege hinweg steigend sind.
	 *
	 * @param raf      Geöffnete Datei
	 * @param fileName Datei
	 *
	 * @param containerIDs hashSet mit ContainerIDs, die in der IndexDatei vorkommen müssen.
	 * @return Ein Defect, falls die Datei Defekt ist, sonst null
	 *
	 * @throws IOException Eingabe-Ausgabe-Fehler
	 */
	private Defect checkDIDIndexFile(final RandomAccessFile raf, File fileName, final HashSet<Long> containerIDs) throws IOException {
		final long length = raf.length();
		long atmax_old = Long.MIN_VALUE;
		long cid_old = Long.MIN_VALUE;

		final HashSet<Long> containerFileIDs = new HashSet<>(containerIDs);

		while(raf.getFilePointer() < length) {
			final long dimin = raf.readLong();
			final long dimax = raf.readLong();
			final long atmin = read6(raf);
			final long atmax = read6(raf);
			final long dtmin = read6(raf);
			final long dtmax = read6(raf);
			final long cid = read5(raf);

			if(!(atmax_old <= atmin)) return new Defect(fileName, _badIndexFiles, "Archivzeit nicht monoton steigend");

			if(!(cid_old < cid)) return new Defect(fileName, _badIndexFiles, "ContainerID nicht streng monoton steigend");

			containerFileIDs.remove(cid);

			atmax_old = atmax;
			cid_old = cid;
		}

		if(containerFileIDs.size() > 0) return new Defect(fileName, _badIndexFiles, "Es kommen nicht alle ContainerIDs in der Indexdatei vor.");

		return null;
	}

	/**
	 * Prüft Indexdateien der Klasse {@link DataTimeIndexImpl}. In dieser Klasse ist gefordert, dass die Einträge nach der minimalen
	 * Datenzeit sortiert sind
	 *
	 * @param raf      Geöffente Datei
	 * @param fileName Datei
	 *
	 * @param containerIDs hashSet mit ContainerIDs, die in der IndexDatei vorkommen müssen.
	 * @return Ein Defect, falls die Datei Defekt ist, sonst null
	 *
	 * @throws IOException Eingabe-Ausgabe-Fehler
	 */
	private Defect checkDataTimeIndexFile(final RandomAccessFile raf, File fileName, final HashSet<Long> containerIDs) throws IOException {
		final long length = raf.length();
		long dtmin_old = Long.MIN_VALUE;

		final HashSet<Long> indexContainerIds = new HashSet<>();
        final HashSet<Long> containerFileIDs = new HashSet<>(containerIDs);

		while(raf.getFilePointer() < length) {
			final long dtmin = read6(raf);
			final long dtmax = read6(raf);
			final long dimin = raf.readLong();
			final long dimax = raf.readLong();
			final long cid = read5(raf);

			if(indexContainerIds.contains(cid)) return new Defect(fileName, _badIndexFiles, "ContainerID mehrfach vorhanden");

			if(!(dtmin_old <= dtmin)) return new Defect(fileName, _badIndexFiles, "Datenzeit nicht monoton steigend");

			indexContainerIds.add(cid);
			containerFileIDs.remove(cid);
			dtmin_old = dtmin;
		}
		if(containerFileIDs.size() > 0) return new Defect(fileName, _badIndexFiles, "Es kommen nicht alle ContainerIDs in der Indexdatei vor.");

		return null;
	}

	/**
	 * Prüft Indexdateien der Klasse {@link DataIndexIndexImpl}. In dieser Klasse ist gefordert, dass die Einträge nach der
	 * minimalen Datenzeit sortiert sind
	 *
	 * @param raf      Geöffente Datei
	 * @param fileName Datei
	 *
     * @param containerIDs hashSet mit ContainerIDs, die in der IndexDatei vorkommen müssen.
	 * @return Ein Defect, falls die Datei Defekt ist, sonst null
	 *
	 * @throws IOException Eingabe-Ausgabe-Fehler
	 */
	private Defect checkDataIndexFile(final RandomAccessFile raf, File fileName, final HashSet<Long> containerIDs) throws IOException {
		final long length = raf.length();
		long dimin_old = Long.MIN_VALUE;

		final HashSet<Long> indexContainerIds = new HashSet<>();
        final HashSet<Long> containerFileIDs = new HashSet<>(containerIDs);

		while(raf.getFilePointer() < length) {
			final long dimin = raf.readLong();
			final long dimax = raf.readLong();
			final long cid = read5(raf);

			if(indexContainerIds.contains(cid)) return new Defect(fileName, _badIndexFiles, "ContainerID mehrfach vorhanden");

			if(!(dimin_old <= dimin)) return new Defect(fileName, _badIndexFiles, "Datenindex nicht monoton steigend");

			indexContainerIds.add(cid);
			containerFileIDs.remove(cid);
			dimin_old = dimin;
		}
		if(containerFileIDs.size() > 0) return new Defect(fileName, _badIndexFiles, "Es kommen nicht alle ContainerIDs in der Indexdatei vor.");

		return null;
	}

	private static long read5(final RandomAccessFile raf) throws IOException {
		return ((raf.readByte() & 0xffL) << 32) + (raf.readInt() & 0xFFFFFFFFL);
	}

	private static long read6(final RandomAccessFile raf) throws IOException {
		return ((raf.readShort() & 0xffffL) << 32) + (raf.readInt() & 0xFFFFFFFFL);
	}

	private static void fixBadIndexFileDefect(final Defect badIndexFileDefect) {
		try {
			File indexFile = badIndexFileDefect.getFile();
			rebuildIndexFile(indexFile);
		}
		catch(Exception e) {
			System.out.println("Fehler beim Wiederherstellen der Indexdatei.");
			e.printStackTrace();
		}
	}

	private static void rebuildIndexFile(final File indexFile) throws IndexException, PersistenceException, SynchronizationFailedException {
		if(indexFile.getName().startsWith("_didx")) {
			rebuildIndexFile(indexFile, "did");
		}
		else if(indexFile.getName().startsWith("_datatime")) {
			rebuildIndexFile(indexFile, "datatime");
		}
		else if(indexFile.getName().startsWith("_dataindex")) {
			rebuildIndexFile(indexFile, "dataindex");
		}
		else if(indexFile.getName().equals(ContainerManagementIndex.IDX_FILENAME)) {
			// Header-Indexe werden vom Archivsystem neu aufgebaut und können hier einfach gelöscht werden
			indexFile.delete();
		}
	}

	private static void rebuildIndexFile(File fileName, String indexKind) throws IndexException, PersistenceException, SynchronizationFailedException {

		String path = fileName.getParentFile().getAbsolutePath();

		String[] fNames = new File(path).list(ContainerFile.CONT_FILENAME_FILTER);
		if(fNames != null && fNames.length > 0) {
			Arrays.sort(fNames);

			ArchiveTimeIndexImpl aTIdx = null;
			DataIndexIndexImpl dIdxIdx = null;
			DataTimeIndexImpl dtidx = null;

			final int REBUILDINDEX_IDX_SIZE = 2048;
			switch(indexKind) {
				case "did":
					if(path.endsWith("oa") || path.endsWith("on")) {
						// Bei aktuellen Daten sind Archivzeit und Datenindex stets monoton
						aTIdx = new DataIndexAndArchiveTimeIndex(REBUILDINDEX_IDX_SIZE, fileName.toPath());
					}
					else {
						// Bei nachgelieferten Daten kann es Datenindex-Rückspruenge geben
						aTIdx = new ArchiveTimeIndexImpl(REBUILDINDEX_IDX_SIZE, fileName.toPath());
					}
					break;
				case "datatime":
					dtidx = new DataTimeIndexImpl(REBUILDINDEX_IDX_SIZE, fileName.toPath());
					break;
				case "dataindex":
					dIdxIdx = new DataIndexIndexImpl(REBUILDINDEX_IDX_SIZE, fileName.toPath());
					break;
			}
			try {
				for(String contFile : fNames) {
					try(BasicContainerFileHandle containerFileReader = new BasicContainerFileHandle(new File(path, contFile))) {
						containerFileReader.ensureHeaderRead();
						long minDidx = containerFileReader.getContainerHeaderParamAsLong(ContainerHdr.CHP_DATA_IDX_MIN);
						long maxDidx = containerFileReader.getContainerHeaderParamAsLong(ContainerHdr.CHP_DATA_IDX_MAX);
						long minDtime = containerFileReader.getContainerHeaderParamAsLong(ContainerHdr.CHP_DATA_TIME_MIN);
						long maxDtime = containerFileReader.getContainerHeaderParamAsLong(ContainerHdr.CHP_DATA_TIME_MAX);
						long minAtime = containerFileReader.getContainerHeaderParamAsLong(ContainerHdr.CHP_ARC_TIME_MIN);
						long maxAtime = containerFileReader.getContainerHeaderParamAsLong(ContainerHdr.CHP_ARC_TIME_MAX);

						boolean rebuild = Long.parseLong(ContainerHdr.CHP_DATA_IDX_MIN.getDefaultVal()) == minDidx
								|| Long.parseLong(ContainerHdr.CHP_DATA_IDX_MAX.getDefaultVal()) == maxDidx
								|| Long.parseLong(ContainerHdr.CHP_DATA_TIME_MIN.getDefaultVal()) == minDtime
								|| Long.parseLong(ContainerHdr.CHP_DATA_TIME_MAX.getDefaultVal()) == maxDtime
								|| Long.parseLong(ContainerHdr.CHP_ARC_TIME_MIN.getDefaultVal()) == minAtime
								|| Long.parseLong(ContainerHdr.CHP_ARC_TIME_MAX.getDefaultVal()) == maxAtime;

						if(!rebuild) {
							try {
								if(dtidx != null) dtidx.add(minDtime, maxDtime, minDidx, maxDidx, containerFileReader.getContainerId());
								if(aTIdx != null) {
									aTIdx.add(minDidx, maxDidx, minAtime, maxAtime, minDtime, maxDtime, containerFileReader.getContainerId());
								}
								if(dIdxIdx != null) dIdxIdx.add(minDidx, maxDidx, containerFileReader.getContainerId());
							}
							catch(IndexException e) {
								rebuild = true;
							}
						}
						if(rebuild) {
							minDidx = Long.MAX_VALUE;
							maxDidx = Long.MIN_VALUE;
							minAtime = Long.MAX_VALUE;
							maxAtime = Long.MIN_VALUE;
							minDtime = Long.MAX_VALUE;
							maxDtime = Long.MIN_VALUE;
							try(DataIterator iter = containerFileReader.iterator()) {
								while(!iter.isEmpty()) {
									long val = iter.peekDataIndex();
									minDidx = val < minDidx ? val : minDidx;
									maxDidx = val > maxDidx ? val : maxDidx;
									val = iter.peekArchiveTime();
									minAtime = val < minAtime ? val : minAtime;
									maxAtime = val > maxAtime ? val : maxAtime;
									val = iter.peekDataTime();
									minDtime = val < minDtime ? val : minDtime;
									maxDtime = val > maxDtime ? val : maxDtime;
									iter.remove();
								}
							}

							if(dtidx != null) dtidx.add(minDtime, maxDtime, minDidx, maxDidx, containerFileReader.getContainerId());
							if(aTIdx != null) aTIdx.add(minDidx, maxDidx, minAtime, maxAtime, minDtime, maxDtime, containerFileReader.getContainerId());
							if(dIdxIdx != null) dIdxIdx.add(minDidx, maxDidx, containerFileReader.getContainerId());
						}
					}
				}
				System.out.println(fileName.toString() + ": Indexdatei wurde repariert");
			}
			finally {
				if(dtidx != null) dtidx.close();
				if(aTIdx != null) aTIdx.close();
				if(dIdxIdx != null) dIdxIdx.close();
			}
		}
		else
		{
			System.out.println(fileName + ": Kann Indexdatei nicht reparieren, Containerdateien sind nicht auffindbar.");
		}
	}

	private void checkContainerFile(final File containerFile) {
		//System.out.println("Prüfe " + containerFile);
		_checkedContainerFileCount++;
		final Defect defect = inspectContainerFile(containerFile);
		if(defect == null) {
			_validContainerFileCount++;
		}
		else {
			_defectContainerFileCount++;
			defect.getCategory().add(defect);
		}
	}

	private Defect inspectContainerFile(final File containerFile) {

		if(!containerFile.exists()) {
			// das kann passieren, nachdem eine Korrektur die datei gelöscht hat und ist kein Fehler.
			return null;
		}

		final long fileLength = containerFile.length();
		if(fileLength < 384) {
			return new Defect(containerFile, _cantFix, "Unerwartete Dateigröße. Erwartet: mindestens 384 Bytes, gefunden: " + fileLength + " Bytes");
		}
		else if(fileLength == 384) {
			return new Defect(containerFile, _shortContainerFiles, "Containerdatei enthält keine Daten");
		}
		try(BufferedInputStream in = new BufferedInputStream(new FileInputStream(containerFile))) {
			byte[] header = new byte[384];
			final int headerLength = in.read(header);
			if(headerLength < 384) {
				return new Defect(
						containerFile, _cantFix, "Unerwartete Anzahl gelesener Headerbytes. Erwartet: 384 Bytes, gefunden: " + headerLength + " Bytes"
				);
			}
			if(header[0] != 'v' || header[1] != '1' || header[2] != '0' || header[3] != '0') {
				return new Defect(
						containerFile,
						_cantFix,
						"Unerwartete Containerversion. Erwartet: 'v100', gefunden: '" + (char) header[0] + (char) header[1] + (char) header[2] + (char) header[3]
								+ "'"
				);
			}
			if(header[4] != 0 || header[5] != 0 || header[6] != 1 || header[7] != 0x78) {
				return new Defect(
						containerFile, _cantFix, "Unerwartete Headerlänge. Erwartet: 0x00000178, gefunden: " + String.format(
						"0x%02x%02x%02x%02x", header[4], header[5], header[6], header[7]
				)
				);
			}
			final char[] deletedParamNameChars = new char[10];
			for(int i = 0; i < deletedParamNameChars.length; i++) {
				deletedParamNameChars[i] = (char) header[0x14d + i];
			}
			final String deletedParamName = String.valueOf(deletedParamNameChars);
			if(!deletedParamName.equals("geloescht=")) {
				return new Defect(containerFile, _cantFix, "Unerwartete Lösch-Kennung. Erwartet: 'geloescht=', gefunden: '" + deletedParamName + "'");
			}
			final boolean containerDeleted;
			if(header[0x14d + 10] == '0') {
				containerDeleted = false;
			}
			else if(header[0x14d + 10] == '1') {
				containerDeleted = true;
			}
			else {
				return new Defect(
						containerFile, _cantFix, "Unerwarteter Wert der Lösch-Kennung. Erwartet: '0' oder '1', gefunden: " + (char) header[0x14d + 10]
				);
			}
			if(containerDeleted) {
				// Hinweis: Der folgende Fehler ist im Archivsystem korrigiert, es können Dateien von älteren Versionen repariert werden:
				//
				// Beim Löschen eines Containers mit der Methode ContainerFile.deleteContainerData() werden die in der Container-Datei enthaltenen Datensätze
				//  ersetzt durch eine optimierte Aufzählung der verwendeten Datensatzindizes. Dabei wird die Methode createDataIndexBuffer() verwendet, die
				//  diese optimierte Aufzählung erzeugt. Diese Methode gibt i.A. ein Byte zuviel zurück. Die deleteContainerData() Methode schreibt die von
				//  createDataIndexBuffer() zurückgelieferten Bytes und verkürzt anschließend die Datei um grundsätzlich ein Byte. In bestimmten Fällen, z.B.,
				//  wenn in der Container-Datei nur ein Datensatz enthalten ist oder nur Datensätze mit Sprüngen im Datensatzindex enthalten sind, liefert die
				//  Methode createDataIndexBuffer() allerdings die korrekte Anzahl Bytes zurück, sodass durch das anschließende Verkürzen der Datei um ein Byte
				//  ein Informationsverlust in der Datei entsteht. Es sollte möglich sein mit Hilfe des maximalen Datensatz-Index, der im Datei-Header
				//  gespeichert ist, das fehlende Byte zu berechnen und die defekte Datei zu reparieren.
				final long mod17remainder = (fileLength - 384) % 17;
				if(mod17remainder != 0) {
					if(mod17remainder == 16) {
						return new Defect(containerFile, _oneByteShortDefects, "Falsche Dateilänge einer gelöschten Datei. Es fehlt ein Byte.");
					}
					else {
						return new Defect(
								containerFile,
								_cantFix,
								"Unerwartete Dateilänge einer gelöschten Datei. Erwartet: 384 + 17 * x, gefunden : 384 + 17 * y + " + mod17remainder
						);
					}
				}

				// Beim Löschen eines Containers, in dem als erster Datensatz ein Potentielle-Lücke-Datensatz enthalten ist, wird die optimierte Aufzählung der
				//  enthaltenen Datensatzindizes mit falschem Endwert (-9223372036854775808) des ersten Bereichs erzeugt. Eine Reparatur der fehlerhaften
				//  Container-Dateien sollte möglich sein, indem der fehlerhafte Endwert -9223372036854775808, durch den entsprechenden Startwert ersetzt wird.
				byte[] range = new byte[17];
				if(in.read(range) == 17) {
					if(range[9] == -128 && range[10] == 0 && range[11] == 0 && range[12] == 0 && range[13] == 0 && range[14] == 0 && range[15] == 0
							&& range[16] == 0) {
						return new Defect(
								containerFile,
								_badRangeEndDefects,
								"Falsche Endangabe eines Datensatzindexbereichs in einem gelöschten Container. Erwartet: gültiger Datensatzindex, gelesen: 0x8000000000000000"
						);
					}
				}
			}
			else {
				// Über Datensätze iterieren
				DataInputStream dataInputStream = new DataInputStream(in);

				while(dataInputStream.available() > 0) {
					try {
						int length = dataInputStream.readInt();

						if(length <= 0) {
							return new Defect(containerFile, _badContainerFileEnd, "Der Datenbereich der Containerdatei ist nicht vollständig geschrieben worden, es befinden sich ungültige Bytes am Dateiende.");
						}

						byte[] tmp = new byte[length];
						dataInputStream.readFully(tmp);

						if(tmp[tmp.length - 3] != '#' || tmp[tmp.length - 2] != '#' || tmp[tmp.length - 1] != '\n') {
							// Jeder Datensatz muss auf "##\n" enden
							return new Defect(containerFile, _badContainerFileEnd, "Der Datenbereich der Containerdatei ist nicht vollständig geschrieben worden, es befinden sich ungültige Bytes am Dateiende.");
						}

					}
					catch(EOFException ignored) {
						return new Defect(containerFile, _badContainerFileEnd, "Der Datenbereich der Containerdatei ist nicht vollständig geschrieben worden, es befinden sich ungültige Bytes am Dateiende.");
					}

				}
			}
		}
		catch(IOException e) {
			e.printStackTrace();
			return new Defect(containerFile, _cantFix, "Lesefehler: " + e);
		}
		// ignored
		return null;
	}

	/**
	 * Beim Löschen eines Containers mit der Methode ContainerFile.deleteContainerData() werden die in der Container-Datei enthaltenen Datensätze ersetzt durch
	 * eine optimierte Aufzählung der verwendeten Datensatzindizes. Dabei wird die Methode createDataIndexBuffer() verwendet, die diese optimierte Aufzählung
	 * erzeugt. Diese Methode gibt i.A. ein Byte zuviel zurück. Die deleteContainerData() Methode schreibt die von createDataIndexBuffer() zurückgelieferten Bytes
	 * und verkürzt anschließend die Datei um grundsätzlich ein Byte. In bestimmten Fällen, z.B., wenn in der Container-Datei nur ein Datensatz enthalten ist oder
	 * nur Datensätze mit Sprüngen im Datensatzindex enthalten sind, liefert die Methode createDataIndexBuffer() allerdings die korrekte Anzahl Bytes zurück,
	 * sodass durch das anschließende Verkürzen der Datei um ein Byte ein Informationsverlust in der Datei entsteht. Eine Reparatur der fehlerhaften
	 * Container-Dateien wird mit dieser Methode durchgeführt, indem mit Hilfe des maximalen Datensatz-Index, der im Datei-Header gespeichert ist, das fehlende
	 * Byte berechnet und die defekte Datei damit korrigiert wird.
	 *
	 * @param defect Identifizierter Defekt einer Containerdatei
	 */
	private static void fixOneByteShortDefect(final Defect defect) {
		try(RandomAccessFile container = new RandomAccessFile(defect.getFile(), "rw")) {
			container.seek(0xA4);
			final byte[] propertyName = new byte[6];
			container.readFully(propertyName);
			final char[] properyValueChars = new char[19];
			for(int i = 0; i < properyValueChars.length; i++) {
				properyValueChars[i] = (char) container.readByte();
			}
			final byte propertySeparator = container.readByte();
			if(propertyName[0] != 'D' || propertyName[1] != 'I' || propertyName[2] != 'm' || propertyName[3] != 'a' || propertyName[4] != 'x'
					|| propertyName[5] != '=') {
				System.out.println(defect.getFile().toString() + ": Reparatur nicht möglich, weil die Kennung DImax= nicht an der erwarteten Stelle ist.");
				return;
			}
			if(propertySeparator != '\n') {
				System.out.println(
						defect.getFile().toString() + ": Reparatur nicht möglich, weil der Wert der Kennung DImax= nicht mit NewLine terminiert ist."
				);
				return;
			}
			final String propertyValueString = String.valueOf(properyValueChars);
			final long propertyValue = Long.parseLong(propertyValueString);
			final long fileLength = container.length();
			container.seek(fileLength - 8);
			final long lastLongValue = container.readLong() << 8;
			if(lastLongValue != Long.MIN_VALUE && lastLongValue != (propertyValue & 0xffffffffffffff00L)) {
				System.out.println(
						defect.getFile().toString() + ": Reparatur nicht möglich, weil die letzten 7 Byte der Datei nicht den erwarteten Werten entsprechen."
				);
				return;
			}
			container.seek(fileLength - 7);
			container.writeLong(propertyValue);
			System.out.println(defect.getFile().toString() + ": Container wurde repariert");
		}
		catch(IOException e) {
			System.out.println(defect.getFile().toString() + ": Unerwarteter Fehler bei der Reparatur: " + e);
		}
	}

	/**
	 * Beim Löschen eines Containers, in dem als erster Datensatz ein Potentielle-Lücke-Datensatz enthalten ist, wird die optimierte Aufzählung der enthaltenen
	 * Datensatzindizes mit falschem Endwert (-9223372036854775808) des ersten Bereichs erzeugt. Eine Reparatur der fehlerhaften Container-Dateien wird durch diese
	 * Methode durchgeführt, indem der fehlerhafte Endwert -9223372036854775808, durch den entsprechenden Startwert ersetzt wird.
	 *
	 * @param defect Identifizierter Defekt einer Containerdatei
	 */
	private static void fixBadRangeEndDefect(final Defect defect) {
		try(RandomAccessFile container = new RandomAccessFile(defect.getFile(), "rw")) {
			container.seek(384);
			final long startRangeIndex = container.readLong();
			final byte separator = container.readByte();
			final long badEndRangeIndex = container.readLong();
			if(separator != '-') {
				System.out.println(defect.getFile().toString() + ": Reparatur nicht möglich, weil Separatorzeichen nicht an erwarteter Stelle.");
				return;
			}
			if(badEndRangeIndex != Long.MIN_VALUE) {
				System.out.println(defect.getFile().toString() + ": Reparatur nicht möglich, weil Bereichsende des ersten Bereichs nicht defekt ist.");
				return;
			}
			container.seek(384 + 9);
			container.writeLong(startRangeIndex);
			System.out.println(defect.getFile().toString() + ": Container wurde repariert");
		}
		catch(IOException e) {
			System.out.println(defect.getFile().toString() + ": Unerwarteter Fehler bei der Reparatur: " + e);
		}
	}

	private void fixBadContainerFileEnd(final Defect defect) {
		File containerFile = defect.getFile();
		long anzDs = -1;
		try(BasicContainerFileHandle handle = new BasicContainerFileHandle(containerFile)) {
			handle.ensureHeaderRead();
			anzDs = handle.getContainerHeaderParamAsLong(ContainerHdr.CHP_ANZ_DS);
		}
		catch(PersistenceException e) {
			System.out.println(containerFile.toString() + ": Unerwarteter Fehler bei der Reparatur: " + e);
		}

		try(RandomAccessFile container = new RandomAccessFile(containerFile, "rw")) {
			container.seek(384);

			long lastValidPosition = 384;
			long actualAnzDs = 0;
			while(container.getFilePointer() < container.length()) {
				try {
					lastValidPosition = container.getFilePointer();
					int length = container.readInt();

					if(length <= 0) {
						break;
					}

					byte[] tmp = new byte[length];
					container.readFully(tmp);

					if(tmp[tmp.length - 3] != '#' || tmp[tmp.length - 2] != '#' || tmp[tmp.length - 1] != '\n') {
						// Jeder Datensatz muss auf "##\n" enden
						break;
					}

				}
				catch(EOFException ignored) {
					break;
				}
				actualAnzDs++;
			}
			if(lastValidPosition == container.length()) {
				System.out.println(containerFile.toString() + ": Unerwarteter Fehler bei der Reparatur: Es wurde kein Fehler zum reparieren gefunden");
			}
			if(anzDs == -1 || anzDs == actualAnzDs) {
				container.setLength(lastValidPosition);
			}
			else {
				System.out.println(containerFile.toString() + ": Fehler bei der Reparatur: Die Anzahl der gültigen Datensätze im Container stimmt nicht mit dem Container-Header überein.");
				return;
			}
		}
		catch(Exception e) {
			System.out.println(containerFile.toString() + ": Unerwarteter Fehler bei der Reparatur: " + e);
			return;
		}
		// ignored

		try {
			rebuildIndexFiles(containerFile.getParentFile());
		}
		catch(Exception e) {
			System.out.println(containerFile.toString() + ": Unerwarteter Fehler bei der Reparatur: " + e);
		}

	}

	private void rebuildIndexFiles(final File parentFile) throws IndexException, PersistenceException, SynchronizationFailedException {
		File[] indexFiles = parentFile.listFiles(_indexFilenameFilter);
		if(indexFiles == null) return;
		for(File indexFile : indexFiles) {
			rebuildIndexFile(indexFile);
		}
	}


	private static class Defect {

		private final File _file;

		private final List<Defect> _category;

		private final String _message;

		private HashSet<Long> _containerIDsInDirectory;

		public Defect(final File file, final List<Defect> category, final String message) {
			_file = file;
			_category = category;
			_message = message;
			System.out.println(file.toString() + ": " + message);
		}


		public List<Defect> getCategory() {
			return _category;
		}

		public File getFile() {
			return _file;
		}

		public String getMessage() {
			return _message;
		}

		public HashSet<Long> getContainerIDsInDirectory() {
			return _containerIDsInDirectory;
		}

		public void setContainerIDsInDirectory(final HashSet<Long> containerIDsInDirectory) {
			_containerIDsInDirectory = containerIDsInDirectory;
		}
	}
}
